import os

import numpy
import pytest
from nabu.stitching.config import PreProcessedZStitchingConfiguration
from nabu.stitching.z_stitching import PreProcessZStitcher
from tomwer.tests.conftest import qtapp  # noqa F401

from silx.image.phantomgenerator import PhantomGenerator

from tomwer.gui.stitching.stitching_preview import PreviewStitchingPlot
from tomwer.gui.stitching.tests.utils import create_scans_z_series


@pytest.mark.parametrize("flip_lr", (True, False))
@pytest.mark.parametrize("flip_ud", (True, False))
def test_preview(tmp_path, flip_lr, flip_ud, qtapp):  # noqa F401
    axis_0_positions = (90, 0.0, -90.0)
    axis_2_positions = (0.0, 0.0, 0.0)
    pixel_size = 1.0

    output_file_path = os.path.join(str(tmp_path), "output", "stitched.nx")
    input_dir = os.path.join(tmp_path, "input")

    scans = create_scans_z_series(
        output_dir=input_dir,
        z_positions_m=axis_0_positions,
        x_positions_m=axis_2_positions,
        shifts=((0.0, 0.0), (-90.0, 0.0), (-180.0, 0.0)),
        pixel_size=pixel_size,
        raw_frame_width=280,
        final_frame_width=100,
        flip_lr=flip_lr,
        flip_ud=flip_ud,
    )
    stitching_config = PreProcessedZStitchingConfiguration(
        output_file_path=output_file_path,
        output_data_path="entry_stitched",
        overwrite_results=True,
        slurm_config=None,
        axis_0_pos_mm=numpy.array(axis_0_positions) / 1000,
        axis_2_pos_mm=numpy.array(axis_2_positions) / 1000,
        axis_0_pos_px=None,
        axis_1_pos_px=None,
        axis_2_pos_px=None,
        input_scans=scans,
        pixel_size=pixel_size,
    )
    widget = PreviewStitchingPlot(axis=0)

    widget._backGroundAction.toggle()

    stitcher = PreProcessZStitcher(configuration=stitching_config)
    stitched_id = stitcher.stitch(store_composition=True)
    assert stitched_id is not None
    composition = stitcher.frame_composition
    assert composition is not None

    widget.setStitchedTomoObj(tomo_obj_id=stitched_id.to_str(), composition=composition)
    assert widget.stitched_image is not None
    assert widget.composition_background is not None

    numpy.testing.assert_almost_equal(
        widget.stitched_image,
        PhantomGenerator.get2DPhantomSheppLogan(n=280).astype(numpy.float32) * 256.0,
    )

    widget._backGroundAction.toggle()



def relative_pos_to_absolute(relative_pos: float, det_width: int):
    """
    Convert relative center of rotation to absolute.
    """
    # Compute the midpoint differently for even and odd widths
    midpoint = det_width / 2.0
    return relative_pos + midpoint


def absolute_pos_to_relative(absolute_pos: float, det_width: int):
    """
    Convert absolute center of rotation to relative.
    """
    # Compute the midpoint differently for even and odd widths
    midpoint = det_width / 2.0
    return absolute_pos - midpoint



# coding: utf-8
from __future__ import annotations

import gc
import logging
import os
import shutil
import tempfile

from silx.gui import qt
from silx.gui.utils.testutils import SignalListener, TestCaseQt

from orangecontrib.tomwer.widgets.control.FilterOW import NameFilterOW
from tomwer.core.scan.edfscan import EDFTomoScan
from tomwer.core.utils.scanutils import MockEDF, MockNXtomo

logging.disable(logging.INFO)


class NameFilterOWCI(NameFilterOW):
    sigScanReady = qt.Signal(str)
    """signal emitted when a scan is ready; Parameter is folder path of the
    scan"""

    def _signalScanReady(self, scan):
        self.sigScanReady.emit(scan.path)


class TestFilterWidget(TestCaseQt):
    """class testing the DarkRefWidget"""

    def setUp(self):
        TestCaseQt.setUp(self)
        self.widget = NameFilterOWCI(parent=None)

        self.tmpdir = tempfile.mkdtemp()
        # create hdf5 dataset
        self._hdf5with_mytotodir = MockNXtomo(
            scan_path=os.path.join(self.tmpdir, "mytotodir120"), n_proj=10
        ).scan
        self._hdf5without_mytotodir = MockNXtomo(
            scan_path=os.path.join(self.tmpdir, "totodir120"), n_proj=10
        ).scan
        # create edf dataset
        mockEDF1 = MockEDF(
            scan_path=os.path.join(self.tmpdir, "mytotodir20"), n_radio=10
        )
        self.edfwith_mytotodir = EDFTomoScan(scan=mockEDF1.scan_path)

        mockEDF2 = MockEDF(scan_path=os.path.join(self.tmpdir, "dir120"), n_radio=10)
        self.edfwithout_mytotodir = EDFTomoScan(scan=mockEDF2.scan_path)

        # add a signal listener
        self.signalListener = SignalListener()
        self.widget.sigScanReady.connect(self.signalListener)

        # define the unix file pattern
        self.widget.setPattern("*mytotodir*")

    def tearDown(self):
        shutil.rmtree(self.tmpdir)
        self.widget.setAttribute(qt.Qt.WA_DeleteOnClose)
        self.widget.close()
        self.widget = None
        gc.collect()

    def testEDF(self):
        """Make sure filtering works with EDF"""
        self.assertEqual(self.signalListener.callCount(), 0)
        self.widget.applyfilter(self.edfwith_mytotodir)
        self.qapp.processEvents()
        self.assertEqual(self.signalListener.callCount(), 1)
        self.widget.applyfilter(self.edfwithout_mytotodir)
        self.qapp.processEvents()
        self.assertEqual(self.signalListener.callCount(), 1)

    def testHDF5(self):
        """Make sure filtering works with HDF5"""
        self.assertEqual(self.signalListener.callCount(), 0)
        self.widget.applyfilter(self._hdf5without_mytotodir)
        self.qapp.processEvents()
        self.assertEqual(self.signalListener.callCount(), 0)
        self.widget.applyfilter(self._hdf5with_mytotodir)
        self.qapp.processEvents()
        self.assertEqual(self.signalListener.callCount(), 1)



from __future__ import annotations

from silx.gui import qt

from tomwer.core.scan.scanbase import TomwerScanBase


class ScanNameLabelAndShape(qt.QWidget):
    """Scan to display the scan name"""

    def __init__(self, parent):
        qt.QWidget.__init__(self, parent=parent)
        self.setLayout(qt.QHBoxLayout())
        self.setContentsMargins(0, 0, 0, 0)
        self.layout().setSpacing(0)

        label = qt.QLabel("scan: ", self)
        label.setSizePolicy(qt.QSizePolicy.Minimum, qt.QSizePolicy.Minimum)
        self.layout().addWidget(label)
        self._scanNameLabel = qt.QLabel("", self)
        self._scanNameLabel.setAlignment(qt.Qt.AlignLeft)
        self._scanNameLabel.setSizePolicy(
            qt.QSizePolicy.Expanding, qt.QSizePolicy.Minimum
        )
        self.layout().addWidget(self._scanNameLabel)

        self._shapeLabel = qt.QLabel("", self)
        self._shapeLabel.setAlignment(qt.Qt.AlignLeft)
        self._shapeLabel.setSizePolicy(qt.QSizePolicy.Expanding, qt.QSizePolicy.Minimum)
        self.layout().addWidget(self._shapeLabel)

        # set up
        self.clear()

    def setScan(self, scan: TomwerScanBase | None):
        if scan is None or scan.path is None:
            self.clear()
        elif not isinstance(scan, TomwerScanBase):
            raise TypeError(
                f"Scan is expected to be an  instance of {TomwerScanBase}. Get {type(scan)} instead"
            )
        else:
            assert isinstance(scan, TomwerScanBase)
            self._scanNameLabel.setText(scan.get_identifier().short_description())
            self._scanNameLabel.setToolTip(scan.get_identifier().to_str())

            shape_x = scan.dim_1 if scan.dim_1 is not None else "?"
            shape_y = scan.dim_2 if scan.dim_2 is not None else "?"
            self._shapeLabel.setText(f"dims: x={shape_x}, y={shape_y}")

    def clear(self):
        self._scanNameLabel.setText("-")
        self._scanNameLabel.setToolTip("")
        self._shapeLabel.setText("")



from __future__ import annotations


import os
import shutil
import tempfile

import numpy
import pytest
from silx.gui import qt
from silx.gui.utils.testutils import TestCaseQt
from tomoscan.esrf.volume.edfvolume import EDFVolume
from tomoscan.esrf.volume.hdf5volume import HDF5Volume

from tomwer.gui.dialog.QVolumeDialog import QVolumeDialog
from tomwer.tests.utils import skip_gui_test


@pytest.mark.skipif(skip_gui_test(), reason="skip gui test")
class TestVolumeDialog(TestCaseQt):
    """Test QVolumeDialog"""

    def setUp(self):
        super().setUp()
        self.tmp_dir = tempfile.mkdtemp()
        self._dialog = QVolumeDialog()

    def tearDown(self):
        shutil.rmtree(self.tmp_dir)
        self._dialog.setAttribute(qt.Qt.WA_DeleteOnClose)
        self._dialog.close()
        self._dialog = None
        return super().tearDown()

    def test_provide_hdf5_volume_auto(self):
        """
        Test everything can be discovered from a single 'default' hdf5 file
        """
        self._dialog.setFullAuto(True)
        file_path = os.path.join(self.tmp_dir, "myvolume.nx")
        volume = HDF5Volume(
            file_path=file_path,
            data_path="entry",
            data=numpy.linspace(start=1, stop=10, num=300, dtype=numpy.uint8).reshape(
                (3, 10, 10)
            ),
        )
        volume.save()
        self._dialog.show()
        self._dialog.setFilePath(file_path)
        self.qapp.processEvents()
        assert self._dialog.getDataPath() in ("entry", "/entry")
        assert self._dialog.getDataExtension() == "nx"
        assert self._dialog.getVolumeBasename() in ("", None)
        volume_defined = self._dialog.getVolume()
        assert volume_defined is not None
        volume_defined.load()
        assert volume_defined.data is not None
        assert (
            self._dialog._volumeUrlQL.text() == volume_defined.get_identifier().to_str()
        )

    def test_provide_hdf5_volume_manual(self):
        """
        Test we can create a volume if everything is defined manually from hdf5 volume file
        """
        self._dialog.setFullAuto(False)
        file_path = os.path.join(self.tmp_dir, "myvolume.nx")
        volume = HDF5Volume(
            file_path=file_path,
            data_path="entry",
            data=numpy.linspace(start=1, stop=10, num=300, dtype=numpy.uint8).reshape(
                (3, 10, 10)
            ),
        )
        volume.save()

        self._dialog.setDataExtension("nx")
        assert self._dialog.getVolume() is not None
        self._dialog.setFilePath(file_path)
        self._dialog.setDataPath("entry")
        volume_defined = self._dialog.getVolume()
        volume_defined.load()
        assert volume_defined.data is not None
        assert (
            self._dialog._volumeUrlQL.text() == volume_defined.get_identifier().to_str()
        )

    def test_provide_edf_volume_auto(self):
        """
        Test everything can be discovered from a single 'default' edf volume folder
        """
        self._dialog.setFullAuto(True)
        volume = EDFVolume(
            folder=self.tmp_dir,
            data=numpy.linspace(start=1, stop=10, num=300, dtype=numpy.uint8).reshape(
                (3, 10, 10)
            ),
        )
        volume.save()
        self._dialog.show()

        self._dialog.setFilePath(self.tmp_dir)
        self.qapp.processEvents()
        assert self._dialog.getDataPath() in ("", None)
        assert self._dialog.getDataExtension() == "edf"
        assert self._dialog.getVolumeBasename() == os.path.basename(self.tmp_dir)
        volume_defined = self._dialog.getVolume()
        assert volume_defined is not None
        volume_defined.load()
        assert volume_defined.data is not None
        assert (
            self._dialog._volumeUrlQL.text() == volume_defined.get_identifier().to_str()
        )

    def test_provide_edf_volume_manual(self):
        """
        Test we can create a volume if everything is defined manually from edf volume folder
        """
        self._dialog.setFullAuto(False)
        volume = EDFVolume(
            folder=self.tmp_dir,
            data=numpy.linspace(start=1, stop=10, num=300, dtype=numpy.uint8).reshape(
                (3, 10, 10)
            ),
            volume_basename="test",
        )
        volume.save()
        self._dialog.show()

        self._dialog.setFilePath(self.tmp_dir)
        self.qapp.processEvents()
        self._dialog.setDataExtension(
            ".edf"
        )  # should be edf but test works also if user add the dot
        self._dialog.setVolumeBasename("test")
        volume_defined = self._dialog.getVolume()
        assert volume_defined is not None
        volume_defined.load()
        assert volume_defined.data is not None
        assert (
            self._dialog._volumeUrlQL.text() == volume_defined.get_identifier().to_str()
        )

    def test_provide_hdf5_volume_url(self):
        """
        Test we can create a volume if the url (of a hdf5 volume) is provided by the user
        """
        file_path = os.path.join(self.tmp_dir, "myvolume.hdf")
        volume = HDF5Volume(
            file_path=file_path,
            data_path="entry",
            data=numpy.linspace(start=1, stop=10, num=300, dtype=numpy.uint8).reshape(
                (3, 10, 10)
            ),
        )
        volume.save()
        self._dialog.setVolumeUrl(volume.get_identifier().to_str())
        volume_defined = self._dialog.getVolume()
        assert volume_defined is not None
        volume_defined.load()
        assert volume_defined.data is not None
        assert self._dialog.getDataPath() in ("entry", "/entry")
        assert self._dialog.getVolumeBasename() in ("", None)
        assert self._dialog.getFilePath() == file_path
        assert self._dialog.getDataExtension() == "hdf"

    def test_provide_edf_volume_url(self):
        """
        Test we can create a volume if the url (of a EDFVolume) is provided by the user
        """
        volume = EDFVolume(
            folder=self.tmp_dir,
            data=numpy.linspace(start=1, stop=10, num=300, dtype=numpy.uint8).reshape(
                (3, 10, 10)
            ),
            volume_basename="test",
        )
        volume.save()
        self._dialog.setVolumeUrl(volume.get_identifier().to_str())
        volume_defined = self._dialog.getVolume()
        assert volume_defined is not None
        volume_defined.load()
        assert volume_defined.data is not None
        assert self._dialog.getDataPath() in ("", None)
        assert self._dialog.getVolumeBasename() == "test"
        assert self._dialog.getFilePath() == self.tmp_dir
        assert self._dialog.getDataExtension() == "edf"



# coding: utf-8
from __future__ import annotations


import gc
import os
import shutil
import tempfile

from silx.gui import qt
from silx.gui.utils.testutils import TestCaseQt

from orangecontrib.tomwer.widgets.visualization.DiffViewerOW import DiffViewerOW
from tomwer.core.utils.scanutils import MockNXtomo


class TestDiffViewerOW(TestCaseQt):
    def setUp(self):
        super().setUp()
        self.tmp_dir = tempfile.mkdtemp()
        os.makedirs(self.tmp_dir, exist_ok=True)
        self.scan = MockNXtomo(
            scan_path=os.path.join(self.tmp_dir, "myscan"),
            n_proj=20,
            n_ini_proj=20,
            dim=10,
        ).scan
        self.widget = DiffViewerOW()

    def tearDown(self):
        self.widget.setAttribute(qt.Qt.WA_DeleteOnClose)
        self.widget.close()
        self.widget = None
        self.qapp.processEvents()
        shutil.rmtree(self.tmp_dir)
        gc.collect()

    def test(self):
        self.widget.addScan(self.scan)



from __future__ import annotations
import functools
from silx.utils.enum import Enum as _Enum
from silx.gui import qt
from tomwer.gui import icons


class LoadingMode(_Enum):
    ON_SHOW_LOADING = "load when show requested"
    LAZY_LOADING = "lazy loading"
    ASAP_LOADING = "load ASAP"

    def icon(self):
        if self is self.ON_SHOW_LOADING:
            return icons.getQIcon("low_speed")
        elif self is self.LAZY_LOADING:
            return icons.getQIcon("medium_low_speed")
        elif self is self.ASAP_LOADING:
            return icons.getQIcon("high_speed")
        else:
            raise ValueError(f"loading mode {self} not handled")

    def tooltip(self):
        if self is self.ON_SHOW_LOADING:
            return "will load data only when requiring"
            return icons.getQIcon("medium_low_speed")
        elif self is self.LAZY_LOADING:
            return "load data with reduce prefetch"
        elif self is self.ASAP_LOADING:
            return "load data as soon as possible"
        else:
            raise ValueError(f"loading mode {self} not handled")


class LoadingModeToolButton(qt.QToolButton):
    """Tool button to switch keep aspect ratio of a plot"""

    sigLoadModeChanged = qt.Signal(str)

    def __init__(self, parent=None):
        super(LoadingModeToolButton, self).__init__(parent=parent)
        menu = qt.QMenu(self)

        action_lazy_loading = LazyLoadingAction(parent=self)
        # by default icons are not visible from the menu. In this case we want to display it
        action_lazy_loading.setIconVisibleInMenu(True)
        fctLL = functools.partial(self.setLoadingMode, LoadingMode.LAZY_LOADING)
        menu.addAction(action_lazy_loading)

        action_asap_loading = ASAP_LoadingAction(parent=self)
        action_asap_loading.setIconVisibleInMenu(True)
        fctAL = functools.partial(self.setLoadingMode, LoadingMode.ASAP_LOADING)
        menu.addAction(action_asap_loading)

        self.setMenu(menu)
        # connect signal / slot
        action_lazy_loading.triggered.connect(fctLL)
        action_asap_loading.triggered.connect(fctAL)

        # set up
        self.setPopupMode(qt.QToolButton.InstantPopup)
        self.setLoadingMode(LoadingMode.LAZY_LOADING)

    def setLoadingMode(self, loadingMode: LoadingMode | str):
        loadingMode = LoadingMode(loadingMode)
        assert loadingMode in LoadingMode
        self.setIcon(loadingMode.icon())
        self.setToolTip(loadingMode.tooltip())
        self.sigLoadModeChanged.emit(loadingMode.value)


class LazyLoadingAction(qt.QAction):
    def __init__(self, parent):
        super().__init__("load data with small prefetch", parent=parent)
        self.setIcon(LoadingMode.LAZY_LOADING.icon())


class ASAP_LoadingAction(qt.QAction):
    def __init__(self, parent):
        super().__init__("load data ASAP", parent=parent)
        self.setIcon(LoadingMode.ASAP_LOADING.icon())



import os
from tomoscan.tests.datasets import GitlabProject

TomwerCIDatasets = GitlabProject(
    branch_name="tomwer",
    host="https://gitlab.esrf.fr",
    cache_dir=os.path.join(
        os.path.dirname(__file__),
        "__archive__",
    ),
    token=None,
    project_id=4299,  # https://gitlab.esrf.fr/tomotools/ci_datasets
)



# coding: utf-8

from silx.utils.enum import Enum as _Enum


class CorAngleMode(_Enum):
    use_0_180 = "0-180"
    use_90_270 = "90-270"
    manual_selection = "manual"

    @classmethod
    def from_value(cls, value):
        # TODO: back compatibility if is int
        if value == 0:
            return CorAngleMode.use_0_180
        elif value == 1:
            return CorAngleMode.use_90_270
        elif value == 2:
            return CorAngleMode.manual_selection
        else:
            return super().from_value(value=value)



# coding: utf-8
from __future__ import annotations

import logging

import tomoscan.esrf.scan.utils
from orangewidget import gui
from orangewidget.settings import Setting
from orangewidget.widget import Input, Output, OWBaseWidget

import tomwer.core.process.control.singletomoobj
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.scan.scanfactory import ScanFactory
from tomwer.core.tomwer_object import TomwerObject
from tomwer.core.volume.volumefactory import VolumeFactory
from tomwer.gui.control.singletomoobj import SingleTomoObj

_logger = logging.getLogger(__name__)


class SingleTomoObjOW(OWBaseWidget, openclass=True):
    name = "single tomo obj"
    id = "orange.widgets.tomwer.control.SingleScanOW.SingleScanOW"
    description = "Definition of a single dataset"
    icon = "icons/single_tomo_obj.svg"
    priority = 51
    keywords = ["tomography", "NXtomo", "tomwer", "folder", "scan"]

    want_main_area = True
    want_control_area = False
    resizing_enabled = True

    ewokstaskclass = tomwer.core.process.control.singletomoobj.SingleTomoObjProcess

    _tomo_obj_setting = Setting(str())

    class Inputs:
        tomo_obj = Input(name="tomo_obj", type=TomwerObject)

    class Outputs:
        tomo_obj = Output(
            name="tomo_obj", type=TomwerObject, doc="one object to be process"
        )
        reduced_darks = Output(
            name="reduced dark(s)",
            type=dict,
            doc="Reduced darks as a dict",
        )
        reduced_flats = Output(
            name="reduced flat(s)",
            type=dict,
            doc="Reduced flats as a dict",
        )
        relative_reduced_darks = Output(
            name="relative reduced dark(s)",
            type=dict,
            doc="Reduced darks as a dict. Indexes are provided as relative so in [0.0, 1.0[",
        )
        relative_reduced_flats = Output(
            name="relative reduced flat(s)",
            type=dict,
            doc="Reduced flats as a dict. Indexes are provided as relative so in [0.0, 1.0[",
        )

    def __init__(self, parent=None):
        super().__init__(parent)
        self._latest_received_scan = None
        # small work around to keep in scan processing cache and avoid recomputing it if not necessary
        self.widget = SingleTomoObj(parent=self)
        self.widget.sigTomoObjChanged.connect(self._updateSettings)
        self.widget.sigTomoObjChanged.connect(self._triggerObjDownstream)

        self._loadSettings()
        layout = gui.vBox(self.mainArea, self.name).layout()
        layout.addWidget(self.widget)

    @Inputs.tomo_obj
    def add(self, tomo_obj):
        if tomo_obj is None:
            return
        self.widget.setTomoObject(tomo_obj)

    def _loadSettings(self):
        if self._tomo_obj_setting != str():
            self.widget.setTomoObject(self._tomo_obj_setting)

    def _updateSettings(self):
        self._tomo_obj_setting = self.widget.getTomoObjIdentifier()

    def _triggerObjDownstream(self):
        obj_identifier = self.widget.getTomoObjIdentifier()
        if (
            self._latest_received_scan is not None
            and self._latest_received_scan.get_identifier() == obj_identifier
        ):
            data = self._latest_received_scan
        else:
            try:
                data = VolumeFactory.create_tomo_object_from_identifier(obj_identifier)
            except:  # noqa E722
                try:
                    data = ScanFactory.create_tomo_object_from_identifier(
                        obj_identifier
                    )
                except:  # noqa E722
                    _logger.warning(
                        f"Unable to find an obj with {obj_identifier} as identifier"
                    )
                    return

        self.Outputs.tomo_obj.send(data)
        # if the data is a scan then provide also access to the reduced frames. Can be convenient
        if isinstance(data, TomwerScanBase):
            if data.reduced_darks not in (None, {}):
                reduced_darks = data.reduced_darks
                self.Outputs.reduced_darks.send(reduced_darks)

                # we want to send those in relative position to have something generic. This is a convention for now
                reduced_darks = (
                    tomoscan.esrf.scan.utils.from_absolute_reduced_frames_to_relative(
                        reduced_frames=data.reduced_darks, scan=data
                    )
                )
                reduced_darks["reduce_frames_name"] = (
                    f"darks from {data.get_identifier().short_description()}"
                )
                self.Outputs.relative_reduced_darks.send(reduced_darks)

            if data.reduced_flats not in (None, {}):
                reduced_flats = data.reduced_flats
                self.Outputs.reduced_flats.send(reduced_flats)

                # we want to send those in relative position to have something generic. This is a convention for now
                reduced_flats = (
                    tomoscan.esrf.scan.utils.from_absolute_reduced_frames_to_relative(
                        reduced_frames=data.reduced_flats, scan=data
                    )
                )
                reduced_flats["reduce_frames_name"] = (
                    f"flats from {data.get_identifier().short_description()}"
                )
                self.Outputs.relative_reduced_flats.send(reduced_flats)



# coding: utf-8

"""
This module is used to define the process of the reference creator.
This is related to the issue #184
"""

from __future__ import annotations


import logging

from silx.gui import qt
from pyunitsystem.metricsystem import MetricSystem

from tomwer.core.utils.char import MU_CHAR

_logger = logging.getLogger(__name__)


class PixelEntry(qt.QWidget):
    valueChanged = qt.Signal()
    """emit when the pixel value change"""

    class Validator(qt.QIntValidator):
        def validate(self, a0: str, a1: int):
            if a0 == "unknow":
                return qt.QValidator.Acceptable
            else:
                return super().validate(a0, a1)

    def __init__(self, name, parent=None):
        qt.QWidget.__init__(self, parent)

        self.setLayout(qt.QHBoxLayout())
        self._label = qt.QLabel(name, parent=self)
        self.layout().addWidget(self._label)
        self._qlePixelSize = qt.QLineEdit(parent=self)
        self._qlePixelSize.setValidator(self.Validator(self._qlePixelSize))
        self._qlePixelSize.setPlaceholderText("px")
        self.layout().addWidget(self._qlePixelSize)

        # connect signal / slot
        self._qlePixelSize.editingFinished.connect(self.valueChanged)

    def getValue(self):
        if self._qlePixelSize.text() in ("unknown", ""):
            return None
        else:
            return int(self._qlePixelSize.text())

    def setValue(self, value: int):
        self._qlePixelSize.setText(str(value))
        self.valueChanged.emit()

    def setReadOnly(self, read_only: bool) -> None:
        self._qlePixelSize.setReadOnly(read_only)


class MetricEntry(qt.QWidget):
    """
    Create a simple line with a name, a QLineEdit and a combobox to define the
    unit in order to define a metric value.

    :param name: name of the metric value to define
    :param default_unit: Default way to present a value when set
    """

    editingFinished = qt.Signal()
    """emit when editing is finished"""

    class DoubleValidator(qt.QDoubleValidator):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.setNotation(qt.QDoubleValidator.ScientificNotation)

        def validate(self, a0: str, a1: int):
            if a0 == "unknown":
                return (qt.QDoubleValidator.Acceptable, a0, a1)
            else:
                return super().validate(a0, a1)

    _CONVERSION = {
        "nm": MetricSystem.NANOMETER.value,
        f"{MU_CHAR}m": MetricSystem.MICROMETER.value,
        "mm": MetricSystem.MILLIMETER.value,
        "cm": MetricSystem.CENTIMETER.value,
        "m": MetricSystem.METER.value,
    }

    valueChanged = qt.Signal()
    """emit when the metric value change"""

    def __init__(self, name, value=0.0, default_unit="m", parent=None):
        qt.QWidget.__init__(self, parent)
        assert type(default_unit) is str
        assert default_unit in ("nm", "mm", "cm", "m", f"{MU_CHAR}m")
        self._base_unit = default_unit

        self.setLayout(qt.QHBoxLayout())
        self._label = qt.QLabel(name, parent=self)
        self.layout().addWidget(self._label)
        self._qlePixelSize = qt.QLineEdit("0.0", parent=self)
        self._qlePixelSize.setValidator(self.DoubleValidator(self._qlePixelSize))
        self.layout().addWidget(self._qlePixelSize)

        self._qcbUnit = qt.QComboBox(parent=self)
        self._qcbUnit.addItem("nm")
        self._qcbUnit.addItem(f"{MU_CHAR}m")
        self._qcbUnit.addItem("mm")
        self._qcbUnit.addItem("cm")
        self._qcbUnit.addItem("m")
        self.layout().addWidget(self._qcbUnit)
        self._resetBaseUnit()

        # connect signal / slot
        self._qcbUnit.currentIndexChanged.connect(self._editingFinished)
        self._qlePixelSize.editingFinished.connect(self._editingFinished)

    def _editingFinished(self, *args, **kwargs):
        self.editingFinished.emit()

    def setReadOnly(self, a0: bool):
        self._qlePixelSize.setReadOnly(a0)
        self._qcbUnit.setEnabled(not a0)

    def setLabelText(self, text: str):
        self._label.setText(text)

    def getCurrentUnit(self):
        assert self._qcbUnit.currentText() in self._CONVERSION
        return self._CONVERSION[self._qcbUnit.currentText()]

    def setValue(self, value_m, displayed_unit: str = "m"):
        """

        :param value: pixel size in international metric system (meter)
        """
        _value = value_m
        if _value in (None, "unknown"):
            txt = "unknown"
        elif isinstance(_value, str):
            if "..." in _value:
                txt = _value
            else:
                try:
                    _value = float(_value)
                except Exception as error:
                    raise ValueError("Given string does not represent a float", error)
                else:
                    assert isinstance(_value, float)
                    txt = str(_value)
        else:
            txt = str(_value)
        self._qlePixelSize.setText(txt)
        self._resetBaseUnit(displayed_unit=displayed_unit)

    def _resetBaseUnit(self, displayed_unit=None):
        """Simple reset of the combobox according to the base_unit"""
        displayed_unit = displayed_unit or self._base_unit
        index = self._qcbUnit.findText(displayed_unit)
        if index is None:
            raise ValueError("unrecognized base unit")
        else:
            self._qcbUnit.setCurrentIndex(index)

    def getValue(self) -> float:
        """

        :return: the value in meter
        """
        if self._qlePixelSize.text() in ("unknown", ""):
            return None
        else:
            return float(self._qlePixelSize.text()) * self.getCurrentUnit()

    def setValidator(self, validator):
        self._qlePixelSize.setValidator(validator)

    def setUnit(self, unit):
        unit = str(MetricSystem.from_value(unit))
        idx = self._qcbUnit.findText(unit)
        self._qcbUnit.setCurrentIndex(idx)



from __future__ import annotations

import logging
import os

import numpy
from nxtomomill.io.config import TomoEDFConfig as EDFConfig
from nxtomomill.io.config import TomoHDF5Config as HDF5Config
from silx.gui import qt

from tomoscan.identifier import BaseIdentifier

from tomwer.gui import icons
from tomwer.gui.qlefilesystem import QLFileSystem
from tomwer.gui.utils.qt_utils import block_signals
from tomwer.io.utils import get_default_directory
from tomwer.core.volume.volumefactory import VolumeFactory
from tomwer.core.process.output import ProcessDataOutputDirMode

from tomwer.core.volume import (
    EDFVolume,
    HDF5Volume,
    JP2KVolume,
    RawVolume,
    TIFFVolume,
    MultiTIFFVolume,
)

_logger = logging.getLogger(__name__)


class SelectionLineEdit(qt.QWidget):
    """Line edit with several type of selection possible:

    * a single value
    * a range of value on the type min:max:step
    * a list of value: val1, val2, ...
    """

    # SINGLE_MODE = 'single'
    RANGE_MODE = "range"
    LIST_MODE = "list"

    # SELECTION_MODES = (SINGLE_MODE, RANGE_MODE, LIST_MODE)
    SELECTION_MODES = (RANGE_MODE, LIST_MODE)

    _DEFAULT_SELECTION = LIST_MODE

    def __init__(self, text=None, parent=None, allow_negative_indices=False):
        qt.QWidget.__init__(self, parent)
        self.setLayout(qt.QHBoxLayout())
        self._qLineEdit = qt.QLineEdit(parent=self)
        if allow_negative_indices:
            fpm = "\\-?\\d*\\.?\\d+"  # float or int matching
        else:
            fpm = "\\d*\\.?\\d+"  # float or int matching
        qRegExp = qt.QRegExp(
            "(" + fpm + "[;]?[,]?[ ]?){1,}" + "|" + ":".join((fpm, fpm, fpm))
        )
        self._qLineEdit.setValidator(qt.QRegExpValidator(qRegExp))
        self.layout().addWidget(self._qLineEdit)
        self._button = SelectionModeButton(parent=self)
        self.layout().addWidget(self._button)

        # QObject signal connections
        self._qLineEdit.textChanged.connect(self._checkIfModeChanged)
        self._button.sigModeChanged.connect(self._modeChanged)

        # expose API
        self.setText = self._qLineEdit.setText
        self.editingFinished = self._qLineEdit.editingFinished
        self.textChanged = self._qLineEdit.textChanged
        self.text = self._qLineEdit.text

        if text is not None:
            self._qLineEdit.setText(str(text))
        # update place holders
        self._modeChanged(self._button.mode)

    def getMode(self):
        return self._button.mode

    @property
    def selection(self):
        if self._qLineEdit.hasAcceptableInput():
            if self._button.mode == self.RANGE_MODE:
                _from, _to, _step = self._qLineEdit.text().split(":")
                _from, _to, _step = float(_from), float(_to), float(_step)
                if _from > _to:
                    _logger.warning(f"to > from, invert {_from} and {_to}")
                    tmp = _to
                    _to = _from
                    _from = tmp
                num = int((_to - _from) / _step)
                return tuple(
                    numpy.linspace(start=_from, stop=_to, num=num, endpoint=True)
                )
            else:
                vals = self._qLineEdit.text().replace(" ", "")
                vals = vals.replace(";", ",").split(",")
                res = []
                [res.append(float(val)) for val in vals]
                if len(res) == 1:
                    return res[0]
                else:
                    return tuple(res)
        else:
            _logger.warning("Wrong input, unvalid selection")
            return None

    def _checkIfModeChanged(self, _str):
        with block_signals(self._button):
            if _str.count(":") > 0:
                self._button.mode = self.RANGE_MODE
            else:
                self._button.mode = self.LIST_MODE

    def _modeChanged(self, mode):
        if mode == self.RANGE_MODE:
            text = "from:to:step"
        elif mode == self.LIST_MODE:
            text = "val1; val2; ..."
        else:
            raise ValueError("unknown mode")

        with block_signals(self._qLineEdit):
            self._qLineEdit.setPlaceholderText(text)


class SelectionModeButton(qt.QToolButton):
    """Base class for Selection QAction.

    :param mode: the mode of selection of the action.
    :param text: The name of this action to be used for menu label
    :param tooltip: The text of the tooltip
    :param triggered: The callback to connect to the action's triggered
                      signal or None for no callback.
    """

    sigModeChanged = qt.Signal(str)
    """signal emit when the mode is changed"""

    def __init__(self, parent=None, tooltip=None, triggered=None):
        qt.QToolButton.__init__(self, parent)
        self._states = {}
        self._mode = None
        for mode in SelectionLineEdit.SELECTION_MODES:
            icon = icons.getQIcon("_".join([mode, "selection"]))
            self._states[mode] = (icon, self._getTooltip(mode))

        self._rangeAction = RangeSelAction(parent=self)
        self._listAction = ListSelAction(parent=self)
        for _action in (self._rangeAction, self._listAction):
            _action.sigModeChanged.connect(self._modeChanged)

        menu = qt.QMenu(self)
        menu.addAction(self._rangeAction)
        menu.addAction(self._listAction)
        self.setMenu(menu)
        self.setPopupMode(qt.QToolButton.InstantPopup)
        self.mode = SelectionLineEdit.LIST_MODE

    def _getTooltip(self, mode):
        # if mode == SelectionLineEdit.SINGLE_MODE:
        #     return 'Define only one value for this parameter'
        if mode == SelectionLineEdit.LIST_MODE:
            return (
                "Define a single value or a list of values for this "
                "parameter (va1; val2)"
            )
        elif mode == SelectionLineEdit.RANGE_MODE:
            return "Define a range of value for this parameter (from:to:step)"
        else:
            raise ValueError("unknow mode")

    def _modeChanged(self, mode):
        self.mode = mode

    @property
    def mode(self):
        return self._mode

    @mode.setter
    def mode(self, mode):
        assert mode in SelectionLineEdit.SELECTION_MODES
        if mode != self._mode:
            self._mode = mode
            self.setIcon(icons.getQIcon("_".join([mode, "selection"])))
            self.setToolTip(self._getTooltip(mode))
            self.sigModeChanged.emit(self._mode)


class SelectionAction(qt.QAction):
    """
    Base class of the several selection mode
    """

    sigModeChanged = qt.Signal(str)
    """emit when the mode change"""

    def __init__(self, mode, parent, text):
        icon = icons.getQIcon("_".join([mode, "selection"]))
        qt.QAction.__init__(self, icon, text, parent)
        self.setIconVisibleInMenu(True)
        self._mode = mode
        self.triggered.connect(self._modeChanged)

    def _modeChanged(self, *args, **kwargs):
        self.sigModeChanged.emit(self._mode)


class RangeSelAction(SelectionAction):
    """
    Action to select a range of element on the scheme from:to:step
    """

    def __init__(self, parent=None):
        SelectionAction.__init__(
            self,
            mode=SelectionLineEdit.RANGE_MODE,
            parent=parent,
            text="range selection",
        )


class ListSelAction(SelectionAction):
    """
    Action to select a list of element on the scheme elmt1, elmt2, ...
    """

    def __init__(self, parent=None):
        SelectionAction.__init__(
            self, mode=SelectionLineEdit.LIST_MODE, parent=parent, text="list selection"
        )


class NXTomomillOutputDirSelector(qt.QWidget):
    sigChanged = qt.Signal()
    """Signal emit when the output directory of the nx file change"""

    def __init__(self, parent=None):
        qt.QWidget.__init__(self, parent)
        self.setLayout(qt.QGridLayout())
        self.__buttonGroup = qt.QButtonGroup(self)
        self.__buttonGroup.setExclusive(True)

        tooltip = """Define the output directory of the nexus (.nx) file. Options are:
        \n - same folder as scan: create the NXtomos at the same level as the bliss input file or spec folder
        \n - 'PROCESSED_DATA' folder: create NXtomos on the default 'PROCESSED_DATA' folder (bliss default folder, nearby the 'raw' folder).
        \n - user defined folder: users can provide their own folders using keywords for string formatting such as 'scan_dir_name', 'scan_basename' or 'scan_parent_dir_basename'
        """

        # output dir is the folder containing the .nx file
        self._inScanFolder = qt.QRadioButton("same folder as scan", self)
        self._inScanFolder.setToolTip(tooltip)
        self.layout().addWidget(self._inScanFolder, 0, 0, 1, 1)
        self.__buttonGroup.addButton(self._inScanFolder)
        # output dir is the default 'reduced'folder
        self._processedDataFolderRB = qt.QRadioButton("'PROCESSED_DATA' folder", self)
        self._processedDataFolderRB.setToolTip(tooltip)
        self.layout().addWidget(self._processedDataFolderRB, 1, 0, 1, 1)
        self.__buttonGroup.addButton(self._processedDataFolderRB)
        # manual
        self._manualRB = qt.QRadioButton("custom output directory", self)
        self._manualRB.setToolTip(tooltip)
        self.layout().addWidget(self._manualRB, 3, 0, 1, 1)
        self._outputFolderQLE = QLFileSystem("", self)
        self.layout().addWidget(self._outputFolderQLE, 3, 1, 1, 1)
        self._selectButton = qt.QPushButton("", self)
        style = qt.QApplication.style()
        icon_opendir = style.standardIcon(qt.QStyle.SP_DirOpenIcon)
        self._selectButton.setIcon(icon_opendir)
        self._selectButton.setToolTip("select output directory")
        self.layout().addWidget(self._selectButton, 3, 2, 1, 1)
        self.__buttonGroup.addButton(self._manualRB)

        # connect signal / slot
        self._selectButton.released.connect(self._selectOutpuFolder)
        self.__buttonGroup.buttonReleased.connect(self._updateVisiblity)
        self._inScanFolder.toggled.connect(self.sigChanged)
        self._processedDataFolderRB.toggled.connect(self.sigChanged)
        self._manualRB.toggled.connect(self.sigChanged)
        self._outputFolderQLE.editingFinished.connect(self.sigChanged)

        # set up
        self._processedDataFolderRB.setChecked(True)
        self._updateVisiblity()

    def _updateVisiblity(self, *args, **kwargs):
        self._selectButton.setVisible(self._manualRB.isChecked())
        self._outputFolderQLE.setVisible(self._manualRB.isChecked())

    def _selectOutpuFolder(self):  # pragma: no cover
        defaultDirectory = self._outputFolderQLE.text()
        if os.path.isdir(defaultDirectory):
            defaultDirectory = get_default_directory()

        dialog = qt.QFileDialog(self, directory=defaultDirectory)
        dialog.setFileMode(qt.QFileDialog.DirectoryOnly)

        if not dialog.exec_():
            dialog.close()
            return

        self._outputFolderQLE.setText(dialog.selectedFiles()[0])
        self.sigChanged.emit()

    def getOutputFolder(self) -> str | ProcessDataOutputDirMode:
        if self._manualRB.isChecked():
            return self._outputFolderQLE.text()
        elif self._processedDataFolderRB.isChecked():
            return ProcessDataOutputDirMode.PROCESSED_DATA_FOLDER
        elif self._inScanFolder.isChecked():
            return ProcessDataOutputDirMode.IN_SCAN_FOLDER
        else:
            raise RuntimeError("Use case - h52nx output dir - not handled")

    def setOutputFolder(self, output_folder: str):
        with block_signals(self):
            self._manualRB.setChecked(output_folder is not None)
            try:
                default_output = ProcessDataOutputDirMode.from_value(output_folder)
            except ValueError:
                self._outputFolderQLE.setText(output_folder)
                self._manualRB.setChecked(True)
            else:
                if default_output is ProcessDataOutputDirMode.IN_SCAN_FOLDER:
                    self._inScanFolder.setChecked(True)
                elif default_output is ProcessDataOutputDirMode.PROCESSED_DATA_FOLDER:
                    self._processedDataFolderRB.setChecked(True)
                else:
                    raise ValueError(f"default output not handled ({default_output})")
            finally:
                self._updateVisiblity()


class _ConfigFileSelector(qt.QWidget):
    """Widget used to select a configuration file. Originally used for
    NXtomomill"""

    sigConfigFileChanged = qt.Signal(str)
    """signal emitted when the edition of the file path is finished"""

    def __init__(self, parent=None, try_load_cfg: bool = True):
        """
        :param try_load_cfg: If True then when a file path is provided will try to load the configuration using '_load_config' and display error if the file is malformed
        """
        super().__init__(parent)
        self._try_load_cfg = try_load_cfg
        self.setLayout(qt.QHBoxLayout())
        self._lineEdit = QLFileSystem("", self)
        self.layout().addWidget(self._lineEdit)
        self._selectButton = qt.QPushButton("select", self)
        self.layout().addWidget(self._selectButton)
        style = qt.QApplication.instance().style()
        icon = style.standardIcon(qt.QStyle.SP_DialogResetButton)
        self._clearButton = qt.QPushButton(self)
        self._clearButton.setIcon(icon)
        self.layout().addWidget(self._clearButton)

        # connect signal / slot button
        self._clearButton.released.connect(self._clearFilePath)
        self._selectButton.released.connect(self._selectCFGFile)
        self._lineEdit.editingFinished.connect(self._editedFinished)

    def _clearFilePath(self):
        self._lineEdit.clear()

    def _selectCFGFile(self):  # pragma: no cover
        dialog = qt.QFileDialog(self)
        dialog.setFileMode(qt.QFileDialog.ExistingFile)

        if not dialog.exec_():
            dialog.close()
            return
        if len(dialog.selectedFiles()) > 0:
            file_path = dialog.selectedFiles()[0]
            self.setCFGFilePath(file_path)

    def _editedFinished(self):
        # try to convert the file and inform the use if this fails
        cfg_file = self.getCFGFilePath()
        if cfg_file not in (None, ""):
            if self._try_load_cfg:
                try:
                    self._load_config_file(cfg_file)
                except Exception as e:
                    mess = f"Fail to load nxtomomill configuration from {cfg_file}. Error is {e}"
                    _logger.warning(mess)
                    qt.QMessageBox.warning(
                        self, "Unable to read configuration from file", mess
                    )
                else:
                    _logger.info(f"Will use {cfg_file} as input configuration file.")
        self.sigConfigFileChanged.emit(cfg_file)

    def getCFGFilePath(self):
        return self._lineEdit.text()

    def setCFGFilePath(self, cfg_file):
        self._lineEdit.setText(cfg_file)
        self._lineEdit.editingFinished.emit()

    def _load_config_file(self, cfg_file: str):
        raise NotImplementedError("Base class")


class HDF5ConfigFileSelector(_ConfigFileSelector):
    def _load_config_file(self, cfg_file: str):
        HDF5Config.from_cfg_file(cfg_file)


class EDFConfigFileSelector(_ConfigFileSelector):
    def _load_config_file(self, cfg_file: str):
        EDFConfig.from_cfg_file(cfg_file)


class OutputVolumeDefinition(qt.QWidget):
    DEFAULT_DATA_PATH = "stitched_volume"

    _OUTPUT_FILE_PATH_TXT = "output file"

    _OUTPUT_FOLDER_PATH_TXT = "output folder"

    def __init__(self, parent=None) -> None:
        super().__init__(parent)
        self.setLayout(qt.QGridLayout())
        # define format
        self._formatLabel = qt.QLabel("format", self)
        self.layout().addWidget(self._formatLabel, 0, 0, 1, 1)
        self._outputFileFormatCB = qt.QComboBox(parent=self)
        self._outputFileFormatCB.addItem("hdf5")
        for volume in (EDFVolume, JP2KVolume, RawVolume, TIFFVolume):
            self._outputFileFormatCB.addItem(volume.DEFAULT_DATA_EXTENSION)
        self.layout().addWidget(self._outputFileFormatCB, 0, 1, 1, 2)

        # output file or folder
        self._outputFileLabel = qt.QLabel("", self)
        self.layout().addWidget(self._outputFileLabel, 1, 0, 1, 1)
        self._outputFileQLE = QLFileSystem("stitched_volume.hdf5", parent=None)
        self.layout().addWidget(self._outputFileQLE, 1, 1, 1, 1)
        self._selectPB = qt.QPushButton("select", self)
        self.layout().addWidget(self._selectPB, 1, 2, 1, 1)

        # output data path
        self._outputDataPathLabel = qt.QLabel("output path", self)
        self.layout().addWidget(self._outputDataPathLabel, 2, 0, 1, 1)
        self._outputDataPathQLE = qt.QLineEdit(self.DEFAULT_DATA_PATH, self)
        self.layout().addWidget(self._outputDataPathQLE, 2, 1, 1, 2)

        # connect signal / path
        self._selectPB.released.connect(self._selectCurrentFileFormatOutputPath)
        self._outputFileFormatCB.currentIndexChanged.connect(self._fileFormatChanged)

        # set up
        self.setOuputVolumeFormat("hdf5")
        # force label update
        self._fileFormatChanged()

    def _fileFormatChanged(self, *args, **kwargs):
        extension = self.getOutputVolumeFormat()
        self._outputDataPathLabel.setVisible(extension == "hdf5")
        self._outputDataPathQLE.setVisible(extension == "hdf5")
        if extension == "hdf5":
            self._outputFileLabel.setText(self._OUTPUT_FILE_PATH_TXT)
        else:
            self._outputFileLabel.setText(self._OUTPUT_FOLDER_PATH_TXT)

    def _selectCurrentFileFormatOutputPath(self):
        extension = self.getOutputVolumeFormat()
        if extension == "hdf5":
            return self.selectOutputFile(
                extensions=[
                    "HDF5 files (*.h5 *.hdf5 *.nx *.nxs *.nexus)",
                    "Nexus files (*.nx *.nxs *.nexus)",
                    "Any files (*)",
                ]
            )
        else:
            return self.selectOutputFolder()

    def selectOutputFile(self, extensions) -> str | None:
        dialog = qt.QFileDialog(self)
        dialog.setNameFilters(extensions)
        if os.environ.get("TOMWER_DEFAULT_INPUT_DIR", None) and os.path.exists(
            os.environ["TOMWER_DEFAULT_INPUT_DIR"]
        ):
            dialog.setDirectory(os.environ["TOMWER_DEFAULT_INPUT_DIR"])
        elif dialog.directory() != os.getcwd() or str(dialog.directory()).startswith(
            "/data"
        ):
            # if the directory as already been set by the user. Avoid redefining it
            pass
        elif os.path.isdir("/data"):
            dialog.setDirectory("/data")

        if not dialog.exec_():
            dialog.close()
            return None
        else:
            return dialog.selectedFiles()

    def selectOutputFolder(self) -> str | None:
        dialog = qt.QFileDialog(self)
        dialog.setFileMode(qt.QFileDialog.DirectoryOnly)
        if os.environ.get("TOMWER_DEFAULT_INPUT_DIR", None) and os.path.exists(
            os.environ["TOMWER_DEFAULT_INPUT_DIR"]
        ):
            dialog.setDirectory(os.environ["TOMWER_DEFAULT_INPUT_DIR"])
        elif dialog.directory() != os.getcwd() or str(dialog.directory()).startswith(
            "/data"
        ):
            # if the directory as already been set by the user. Avoid redefining it
            pass
        elif os.path.isdir("/data"):
            dialog.setDirectory("/data")

        if not dialog.exec_():
            dialog.close()
            return

        return dialog.selectedFiles()

    def getOutputVolumeFormat(self) -> str:
        return self._outputFileFormatCB.currentText()

    def setOuputVolumeFormat(self, file_format: str):
        idx = self._outputFileFormatCB.findText(file_format)
        if idx >= 0:
            self._outputFileFormatCB.setCurrentIndex(idx)

    def getOutputDataPath(self) -> str:
        return self._outputDataPathQLE.text()

    def setOutputDataPath(self, data_path: str):
        self._outputDataPathQLE.setText(data_path)

    def getOutputFilePath(self) -> str:
        return self._outputFileQLE.text()

    def setOutputFilePath(self, file_path: str):
        self._outputFileQLE.setText(file_path)

    def getOutputVolumeIdentifier(self) -> BaseIdentifier:
        volume_format = self.getOutputVolumeFormat()
        file_path = self.getOutputFilePath()

        if volume_format in ("hdf5", "h5", "hdf", "nx"):
            data_path = self.getOutputDataPath()
            if data_path == "":
                data_path = self.DEFAULT_DATA_PATH
            return HDF5Volume(
                file_path=file_path,
                data_path=data_path,
            ).get_identifier()
        elif volume_format == EDFVolume.DEFAULT_DATA_EXTENSION:
            return EDFVolume(
                folder=file_path,
            ).get_identifier()
        elif volume_format == JP2KVolume.DEFAULT_DATA_EXTENSION:
            return JP2KVolume(
                folder=file_path,
            ).get_identifier()
        elif volume_format == RawVolume.DEFAULT_DATA_EXTENSION:
            return RawVolume(
                file_path=file_path,
            ).get_identifier()
        elif volume_format == TIFFVolume.DEFAULT_DATA_EXTENSION:
            return TIFFVolume(
                folder=file_path,
            ).get_identifier()
        else:
            raise RuntimeError(f"format {volume_format} is not handled")

    def setOutputVolumeIdentifier(self, url):
        volume = VolumeFactory.create_tomo_object_from_identifier(url)

        file_format = (
            "hdf5" if isinstance(volume, HDF5Volume) else volume.DEFAULT_DATA_EXTENSION
        )
        self.setOuputVolumeFormat(file_format=file_format)
        if isinstance(volume, HDF5Volume):
            self.setOutputDataPath(volume.data_path)
            self.setOutputFilePath(volume.file_path)
        elif isinstance(
            volume, (EDFVolume, TIFFVolume, JP2KVolume, RawVolume, MultiTIFFVolume)
        ):
            self.setOutputFilePath(volume.data_url.file_path())
        else:
            raise ValueError(f"volume type ({type(volume)}) is not handled")



# coding: utf-8
from __future__ import annotations

import logging

from orangewidget import gui, widget
from orangewidget.widget import Input

from silx.gui import qt

import tomwer.core.process.visualization.diffviewer
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.gui.visualization.diffviewer import DiffFrameViewer

_logger = logging.getLogger(__name__)


class DiffViewerOW(widget.OWBaseWidget, openclass=True):
    """
    Associate TomoScanBase with the silx's ComparaImage tool.
    Allows to compare two random frame.
    """

    name = "diff frame viewer"
    id = "orangecontrib.tomwer.widgets.visualization.diffviewerow"
    description = "Allows comparison between two random frame from a scan"
    icon = "icons/diff.png"
    priority = 107
    keywords = ["tomography", "diff", "tomwer", "compare", "comparison"]

    want_main_area = True
    want_control_area = False
    resizing_enabled = True

    ewokstaskclass = tomwer.core.process.visualization.diffviewer._DiffViewerPlaceHolder

    class Inputs:
        data = Input(name="data", type=TomwerScanBase)

    def __init__(self, parent=None):
        widget.OWBaseWidget.__init__(self, parent)
        self._layout = gui.vBox(self.mainArea, self.name).layout()
        self.viewer = DiffFrameViewer(parent=self)
        self._layout.addWidget(self.viewer)

    @Inputs.data
    def addScan(self, scan):
        if scan is None:
            return
        self.viewer.addScan(scan)

    def sizeHint(self):
        return qt.QSize(500, 700)



"""contains dialogs to select a Volume (reconstruction of a scan)
"""

from __future__ import annotations

import logging
import os

import h5py
from silx.gui import qt
from silx.gui.dialog.DataFileDialog import DataFileDialog
from silx.gui.utils import blockSignals
from silx.io.url import DataUrl
from tomoscan.esrf.volume.utils import (
    get_most_common_extension as get_most_common_extension,
)
from tomoscan.esrf.volume.utils import guess_volumes as tomoscan_guess_volumes
from tomoscan.volumebase import VolumeBase

from tomwer.core.volume import (
    EDFVolume,
    HDF5Volume,
    JP2KVolume,
    MultiTIFFVolume,
    RawVolume,
    TIFFVolume,
)
from tomwer.core.volume.volumefactory import VolumeFactory as Factory
from tomwer.io.utils.tomoobj import DEFAULT_SCHEME_TO_VOL
from tomwer.gui.qlefilesystem import QLFileSystem
from .QDataDialog import QDataDialog


_logger = logging.getLogger()


class QVolumeDialog(qt.QDialog):
    """
    dialog to select / define a volume
    """

    _EDF_EXTENSIONS = ("edf",)
    _TIFF_EXTENSIONS = (
        "tiff",
        "tif",
    )
    _JP2K_EXTENSIONS = (
        "jp2",
        "jp2k",
    )
    _HDF5_EXTENSIONS = ("hdf", "hdf5", "h5", "nx", "nexus")
    _RAW_EXTENSIONS = (
        "vol",
        "raw",
    )

    _VOLUME_TO_EXT = {
        EDFVolume: "edf",
        HDF5Volume: "hdf5",
        TIFFVolume: "tiff",
        MultiTIFFVolume: "tiff",
        JP2KVolume: "jp2",
        RawVolume: "vol",
    }

    def __init__(self, parent=None) -> None:
        super().__init__(parent)
        self.setLayout(qt.QGridLayout())

        # add volume url
        self.layout().addWidget(qt.QLabel("volume url", parent=self), 0, 0, 1, 1)
        self._volumeUrlQL = qt.QLineEdit(parent=self)
        self._volumeUrlQL.setPlaceholderText("scheme:obj_type:path?queries")
        self._volumeUrlQL.setToolTip(
            "volume url defining the volume. Will be automatically updated from the other widgets"
        )
        self.layout().addWidget(self._volumeUrlQL, 0, 1, 1, 3)
        # add file path
        self.layout().addWidget(
            qt.QLabel("file or folder path", parent=self), 1, 0, 1, 1
        )
        self._filePathQL = QLFileSystem(text="", parent=self)
        self._filePathQL.setToolTip(
            "File path (for hdf5 and multi-tiff) or folder path (for edf, jp2k, tiff...)"
        )
        self.layout().addWidget(self._filePathQL, 1, 1, 1, 2)
        self._selectPathPB = qt.QPushButton("select", self)
        self.layout().addWidget(self._selectPathPB, 1, 3, 1, 1)

        # data path
        self.layout().addWidget(
            qt.QLabel("data path", parent=self),
            3,
            0,
            1,
            1,
        )
        self._dataPathQL = qt.QLineEdit(parent=self)
        self._dataPathQL.setToolTip("data path for hdf5 file")
        self.layout().addWidget(self._dataPathQL, 3, 1, 1, 1)
        self._guessDataPathAuto = qt.QCheckBox("guess auto")
        self.layout().addWidget(self._guessDataPathAuto, 3, 2, 1, 1)
        self._guessDataPathAuto.setToolTip(
            "If checked will try to guess automatically data path from given hdf5 file"
        )
        self._selectDataPathPB = qt.QPushButton("select", self)
        self.layout().addWidget(self._selectDataPathPB, 3, 3, 1, 1)

        # basename
        self.layout().addWidget(
            qt.QLabel("basename", parent=self),
            4,
            0,
            1,
            1,
        )
        self._basenameQL = qt.QLineEdit(parent=self)
        self._basenameQL.setToolTip(
            "volume basename. Used for edf, tiff and jp2k volumes"
        )
        self.layout().addWidget(self._basenameQL, 4, 1, 1, 1)
        self._guessBasenameAuto = qt.QCheckBox("guess auto")
        self.layout().addWidget(self._guessBasenameAuto, 4, 2, 1, 1)
        self._guessBasenameAuto.setToolTip(
            "If checked will try to guess automatically basename from given folder"
        )
        # file extension
        self.layout().addWidget(
            qt.QLabel("extension", parent=self),
            5,
            0,
            1,
            1,
        )
        self._extensionQL = qt.QLineEdit(parent=self)
        self._extensionQL.setToolTip("file extension")
        self.layout().addWidget(self._extensionQL, 5, 1, 1, 1)
        self._guessExtensionAuto = qt.QCheckBox("guess auto")
        self.layout().addWidget(self._guessExtensionAuto, 5, 2, 1, 1)
        self._guessExtensionAuto.setToolTip(
            "If checked will try to guess automatically basename from given folder"
        )

        examples_text = """
        --- extra informations. ---

        This dialog has been though to be work with "automatic guess" active. It should work if you didn't used
        any 'fancy' volumes definition. In this case providing the file path of folder path should be enought.
        Regarding the 'fancy' definition be aware that: \n

           - for an hdf5 file: expects hdf5 file path and a data_path. \n
           - for multitiff file: expects the file path only \n
           - for 'folder' volumes (edf, tiff and jp2k): expects folder, basename (default is the folder name) \n
        """
        self._examplesQLE = qt.QTextEdit(self)
        self._examplesQLE.setText(examples_text)
        self._examplesQLE.setReadOnly(True)
        self.layout().addWidget(self._examplesQLE, 99, 0, 4, 4)

        # buttons for validation
        self._buttons = qt.QDialogButtonBox(self)
        types = qt.QDialogButtonBox.Ok | qt.QDialogButtonBox.Cancel
        self._buttons.setStandardButtons(types)

        self.layout().addWidget(self._buttons, 200, 0, 1, 4)

        # connect signal / slot
        self._buttons.button(qt.QDialogButtonBox.Ok).clicked.connect(self.accept)
        self._buttons.button(qt.QDialogButtonBox.Cancel).clicked.connect(self.reject)
        self._guessBasenameAuto.toggled.connect(self._basenameQL.setDisabled)
        self._guessDataPathAuto.toggled.connect(self._dataPathQL.setDisabled)
        self._guessExtensionAuto.toggled.connect(self._extensionQL.setDisabled)
        self._volumeUrlQL.textChanged.connect(self._volumeUrlChanged)
        self._filePathQL.textChanged.connect(self._filePathChanged)
        self._dataPathQL.textChanged.connect(self._dataChanged)
        self._basenameQL.textChanged.connect(self._basenameChanged)
        self._extensionQL.textChanged.connect(self._extensionChanged)
        self._selectPathPB.released.connect(self._selectFileOrFolder)
        self._selectDataPathPB.released.connect(self._selectDataPath)

        # set up - by default try to guess most of it automatically
        self._guessBasenameAuto.setChecked(True)
        self._guessDataPathAuto.setChecked(True)
        self._guessExtensionAuto.setChecked(True)

    def _selectFileOrFolder(self):
        dialog = QDataDialog(self, multiSelection=True)
        # AnyFile because for HDF5 a file is expected, for .edf... a folder is expected
        dialog.setFileMode(qt.QFileDialog.AnyFile)
        dialog.setNameFilters(
            [
                "Any files (*)",
                "HDF5 files (*.h5 *.hdf5 *.hdf *.nx *.nxs *.nexus)",
                "Nexus files (*.nx *.nxs *.nexus)",
                "multitiff files (*.tiff *.tif)",
                "raw files (*.vol *.raw)",
            ]
        )
        if dialog.exec_() == qt.QDialog.Accepted:
            selected_files = dialog.selectedFiles()
            if len(selected_files) > 0:
                file_or_folder = selected_files[0]
                # in case the user provided a .edf, .tiff... file try to make is life simpler.
                if file_or_folder.endswith(
                    (
                        f".{EDFVolume.DEFAULT_DATA_EXTENSION}",
                        f".{TIFFVolume.DEFAULT_DATA_EXTENSION}",
                        f".{JP2KVolume.DEFAULT_DATA_EXTENSION}",
                        f".{JP2KVolume.DEFAULT_DATA_EXTENSION}",
                    )
                ):
                    file_or_folder = os.path.dirname(file_or_folder)

                self.setFilePath(file_or_folder)

    def _selectDataPath(self):
        self._guessDataPathAuto.setChecked(False)
        file_path = self.getFilePath()
        if file_path in (None, "") or not h5py.is_hdf5(file_path):
            _logger.warning(
                f"{file_path} is not an hdf5 file. can't browse it to define a data path"
            )
            return
        dialog = DataFileDialog(self)
        dialog.selectFile(file_path)

        if dialog.exec_():
            try:
                url = dialog.selectedUrl()
                url = DataUrl(path=url)
                if url is not None:
                    if url.file_path() != self.getFilePath():
                        # in case user get data path from another file update it too
                        with blockSignals(self._filePathQL):
                            self.setFilePath(url.file_path())
                        with blockSignals(self._dataPathQL):
                            self.setDataPath(url.data_path())
                    self._updateVolume_url()
            except Exception as e:
                _logger.error(e)

    def _volumeUrlChanged(self):
        url = self._volumeUrlQL.text()
        try:
            volume = Factory.create_tomo_object_from_identifier(url)
        except Exception as e:
            _logger.info(f"fail to create a volume from {url}. Error is {e}")
        else:
            try:
                if isinstance(volume, HDF5Volume):
                    data_path = volume.url.data_path()
                    file_path = volume.file_path
                    volume_basename = ""
                    volume_extension = os.path.splitext(volume.file_path)[-1].lstrip(
                        "."
                    )
                    # handle case no file path provided but only the extension
                    if volume_extension == "":
                        volume_extension = None
                elif isinstance(volume, MultiTIFFVolume):
                    data_path = ""
                    file_path = volume.file_path
                    volume_basename = ""
                    if volume.file_path is not None:
                        volume_extension = os.path.splitext(volume.file_path)[
                            -1
                        ].lstrip(".")
                    else:
                        volume_extension = None
                elif isinstance(volume, (EDFVolume, TIFFVolume, JP2KVolume)):
                    data_path = ""
                    file_path = volume.url.file_path()
                    volume_basename = volume.get_volume_basename()
                    volume_extension = volume.data_extension
                elif isinstance(volume, RawVolume):
                    data_path = ""
                    file_path = volume.url.file_path()
                    volume_basename = ""
                    volume_extension = os.path.splitext(volume.file_path)[-1].lstrip(
                        "."
                    )
                else:
                    raise TypeError(f"volume type {type(volume)} is not handled")
            except (AttributeError, ValueError) as e:
                _logger.error(
                    f"fail to update volume interface from url {url}. Error is {e}"
                )

            if volume_extension is not None:
                with blockSignals(self._extensionQL):
                    self.setDataExtension(volume_extension)
                assert url == self._volumeUrlQL.text()

            with blockSignals(self._filePathQL):
                self.setFilePath(file_path)
                assert url == self._volumeUrlQL.text()
            with blockSignals(self._dataPathQL):
                self.setDataPath(data_path)
                assert url == self._volumeUrlQL.text()

            with blockSignals(self._basenameQL):
                self.setVolumeBasename(volume_basename)
                assert url == self._volumeUrlQL.text()

    def _updateVolume_url(self, *args, **kwargs):
        volume = self.getVolume()
        if volume is None:
            return
        else:
            identifier = volume.get_identifier().to_str()
            with blockSignals(self._volumeUrlQL):
                self._volumeUrlQL.setText(identifier)

    def _filePathChanged(self, *args, **kwargs):
        guess_basename = self._guessBasenameAuto.isChecked()
        guess_data_path = self._guessDataPathAuto.isChecked()
        guess_extension = self._guessExtensionAuto.isChecked()

        file_path = self.getFilePath()
        try_guess_volume = (
            guess_basename or guess_data_path or guess_extension
        ) and os.path.exists(file_path)
        if try_guess_volume:
            guessed_volumes = tomoscan_guess_volumes(
                file_path,
                scheme_to_vol=DEFAULT_SCHEME_TO_VOL,
                filter_histograms=True,
            )

            if guessed_volumes is None or len(guessed_volumes) == 0:
                data_path = ""
                data_extension = ""
                basename = ""
            else:
                first_vol = guessed_volumes[0]
                if len(guessed_volumes) > 1:
                    _logger.warning(
                        f"More than one volume found in {file_path}. Will only take the first one."
                    )
                if isinstance(first_vol, HDF5Volume):
                    data_path = first_vol.url.data_path()
                else:
                    data_path = ""

                if os.path.isfile(file_path):
                    _, data_extension = os.path.splitext(file_path)
                    if data_extension == "":
                        # in the folder use case provide a valid extension
                        data_extension = self._VOLUME_TO_EXT[type(first_vol)]
                    else:
                        data_extension = data_extension.lstrip(".")
                else:
                    data_extension = get_most_common_extension(file_path) or ""

                if isinstance(first_vol, (HDF5Volume, MultiTIFFVolume, RawVolume)):
                    basename = ""
                elif isinstance(first_vol, (EDFVolume, TIFFVolume, JP2KVolume)):
                    basename = os.path.basename(first_vol.get_volume_basename())
                else:
                    raise NotImplementedError
            if guess_extension:
                with blockSignals(self._extensionQL):
                    self._extensionQL.setText(data_extension)
            if guess_basename:
                with blockSignals(self._basenameQL):
                    self._basenameQL.setText(basename)
            if guess_data_path:
                with blockSignals(self._dataPathQL):
                    self._dataPathQL.setText(data_path)
            self._updateVolume_url()

    def _dataChanged(self, *args, **kwargs):
        self._updateVolume_url()

    def _basenameChanged(self, *args, **kwargs):
        self._updateVolume_url()

    def _extensionChanged(self, *args, **kwargs):
        self._updateVolume_url()

    def setFullAuto(self, auto: bool):
        for widget in (
            self._guessBasenameAuto,
            self._guessDataPathAuto,
            self._guessExtensionAuto,
        ):
            widget.setChecked(auto)

    def getFilePath(self):
        return self._filePathQL.text()

    def setFilePath(self, text):
        self._filePathQL.setText(text)

    def getDataPath(self):
        return self._dataPathQL.text()

    def setDataPath(self, text):
        self._dataPathQL.setText(text)

    def getVolumeBasename(self):
        return self._basenameQL.text()

    def setVolumeBasename(self, text: str):
        self._basenameQL.setText(text)

    def getDataExtension(self):
        return self._extensionQL.text()

    def setDataExtension(self, text):
        self._extensionQL.setText(text)

    def setVolumeUrl(self, text: str):
        self._volumeUrlQL.setText(text)

    def getVolume(self) -> VolumeBase:
        """
        Return possible volume from current information.
        this function is also used to determine the volume url

        :warning: the return volume might not exists (at all).
        """
        data_path = self._dataPathQL.text()
        file_path = self._filePathQL.text()
        basename = self._basenameQL.text()
        file_extension = self._extensionQL.text().lower().lstrip(".")

        single_frame_extensions = (
            list(self._EDF_EXTENSIONS)
            + list(self._TIFF_EXTENSIONS)
            + list(self._JP2K_EXTENSIONS)
        )
        if file_extension in self._TIFF_EXTENSIONS and os.path.isfile(file_path):
            for name, var in {"basename": basename, "data path": data_path}.items():
                if var not in ("", None):
                    _logger.warning(
                        f"{name} provided but not used for {MultiTIFFVolume}. Will be ignored"
                    )
            return MultiTIFFVolume(file_path=file_path)

        if file_extension in single_frame_extensions:
            if file_extension in self._EDF_EXTENSIONS:
                constructor = EDFVolume
            elif file_extension in self._TIFF_EXTENSIONS:
                constructor = TIFFVolume
            elif file_extension in self._JP2K_EXTENSIONS:
                constructor = JP2KVolume
            else:
                raise NotImplementedError(
                    f"unhandled file extension ({file_extension})"
                )
            volume = constructor(
                folder=file_path,
                volume_basename=None if basename in ("", None) else basename,
                data_extension=self._extensionQL.text().lstrip("."),
            )
            if data_path not in (None, ""):
                _logger.warning(
                    f"{constructor} are not handling 'data path'. This information will be ignored"
                )
            return volume
        elif file_extension in self._HDF5_EXTENSIONS:
            if basename not in ("", None):
                _logger.warning(
                    f"basename provided but not used for {HDF5Volume}. Will be ignored"
                )
            return HDF5Volume(
                file_path=file_path,
                data_path=data_path,
            )
        elif file_extension in self._RAW_EXTENSIONS:
            return RawVolume(file_path=file_path)
        elif file_extension in (None, ""):
            return None



# coding: utf-8
from __future__ import annotations


import gc
import os
import pickle
import shutil
import tempfile

from orangecanvas.scheme.readwrite import literal_dumps
from silx.gui import qt
from silx.gui.utils.testutils import TestCaseQt

from orangecontrib.tomwer.widgets.control.DataSelectorOW import DataSelectorOW
from tomwer.core.utils.scanutils import MockNXtomo
from tomwer.synctools.axis import QAxisRP


class TestDataSelectorOw(TestCaseQt):
    def setUp(self):
        super().setUp()
        self.tempdir = tempfile.mkdtemp()
        dim = 10
        mock = MockNXtomo(
            scan_path=os.path.join(self.tempdir, "scan1"),
            n_proj=10,
            n_ini_proj=10,
            scan_range=180,
            dim=dim,
        )
        self.scan1 = mock.scan
        mock = MockNXtomo(
            scan_path=os.path.join(self.tempdir, "scan2"),
            n_proj=10,
            n_ini_proj=10,
            scan_range=180,
            dim=dim,
        )
        self.scan2 = mock.scan
        self.widget = DataSelectorOW()

    def tearDown(self):
        self.widget.setAttribute(qt.Qt.WA_DeleteOnClose)
        self.widget.close()
        self.widget = None
        shutil.rmtree(self.tempdir)
        gc.collect()

    def test(self):
        for scan in (self.scan1, self.scan2):
            assert scan.axis_params is None
            scan.axis_params = QAxisRP()
            self.widget.add(scan)
            self.qapp.processEvents()
            scan_1_identifier = scan.get_identifier()
            assert scan_1_identifier.to_str() in self.widget.widget.dataList._myitems
            item = self.widget.widget.dataList._myitems[scan_1_identifier.to_str()]
            data_obj = item.data(qt.Qt.UserRole)
            assert id(scan) == id(data_obj)
            assert scan.axis_params is not None
        self.widget.selectAll()
        self.widget.send()
        self.qapp.processEvents()

        # test serialization
        assert len(self.widget._scanIDs) == 2
        pickle.dumps(self.widget._scanIDs)
        literal_dumps(self.widget._scanIDs)



import gc
import pickle
import numpy

from silx.gui import qt
from silx.gui.utils.testutils import TestCaseQt

from orangecontrib.tomwer.widgets.control.ReduceDarkFlatSelectorOW import (
    ReduceDarkFlatSelectorOW,
)
from orangecanvas.scheme.readwrite import literal_dumps


class TestEmailOW(TestCaseQt):
    def setUp(self):
        super().setUp()
        self.widget = ReduceDarkFlatSelectorOW()
        self.widget._dialog.addReduceFrames(
            {
                "reduce_frames_name": "new dict",
                -1: numpy.random.random(100 * 100).reshape(100, 100),
                2: numpy.random.random(100 * 100).reshape(100, 100),
            },
            selected=(-1, 2),
        )

    def tearDown(self):
        self.widget.setAttribute(qt.Qt.WA_DeleteOnClose)
        self.widget.close()
        self.widget = None
        gc.collect()

    def test(self):
        self.widget.show()

    def test_serializing(self):
        pickle.dumps(self.widget.getConfiguration())

    def test_literal_dumps(self):
        literal_dumps(self.widget.getConfiguration())



# coding: utf-8
from __future__ import annotations


import os
import shutil
import tempfile

import h5py
import numpy
import pytest
from silx.gui import qt
from silx.gui.utils.testutils import TestCaseQt
from silx.io.url import DataUrl

from tomwer.core.process.reconstruction.sadeltabeta.params import SADeltaBetaParams
from tomwer.core.process.reconstruction.scores import ComputedScore
from tomwer.core.process.reconstruction.scores.params import ScoreMethod
from tomwer.core.utils.scanutils import MockNXtomo
from tomwer.gui.reconstruction.sadeltabeta import SADeltaBetaWindow
from tomwer.tests.utils import skip_gui_test


@pytest.mark.skipif(skip_gui_test(), reason="skip gui test")
class TestSADeltaBetaWindow(TestCaseQt):
    """Test the SAAxisWindow interface"""

    _N_OUTPUT_URLS = 10

    def setUp(self):
        super().setUp()
        self._window = SADeltaBetaWindow()
        self.folder = tempfile.mkdtemp()
        self._output_urls = []
        self._cor_scores = {}

        for i in range(self._N_OUTPUT_URLS):
            output_file = os.path.join(self.folder, f"recons_{i}.h5")
            with h5py.File(output_file, mode="a") as h5f:
                h5f["data"] = numpy.random.random(100 * 100).reshape(100, 100)
                url = DataUrl(file_path=output_file, data_path="data", scheme="silx")
                assert url.is_valid()
            self._output_urls.append(url)
            score = ComputedScore(
                tv=numpy.random.randint(10),
                std=numpy.random.randint(100),
            )
            self._cor_scores[i] = (url, score)

        # create a scan
        self.folder = tempfile.mkdtemp()
        dim = 10
        mock = MockNXtomo(
            scan_path=self.folder, n_proj=10, n_ini_proj=10, scan_range=180, dim=dim
        )
        mock.add_alignment_radio(index=10, angle=90)
        mock.add_alignment_radio(index=10, angle=0)
        self.scan = mock.scan

        self._window.setScan(self.scan)

    def tearDown(self):
        self._window.setAttribute(qt.Qt.WA_DeleteOnClose)
        self._window.close()
        self._window = None
        shutil.rmtree(self.folder)
        super().tearDown()

    def testGetConfiguration(self):
        """Insure configuration getter works"""
        conf = self._window.getConfiguration()
        # the configuration should be convertible to
        conf = SADeltaBetaParams.from_dict(conf)
        self.assertEqual(conf.slice_indexes, "middle")
        self.assertTrue(isinstance(conf.nabu_params, dict))
        for key in (
            "preproc",
            "reconstruction",
            "dataset",
            "tomwer_slices",
            "output",
            "phase",
        ):
            with self.subTest(key=key):
                self.assertTrue(key in conf.nabu_params)
        self.assertEqual(conf._dry_run, False)
        self.assertEqual(conf.score_method, ScoreMethod.TV_INVERSE)
        self.assertEqual(conf.output_dir, None)
        self.assertEqual(conf.scores, None)

    def testSetConfiguration(self):
        """Insure configuration setter works"""
        nabu_specific_params = {
            "phase": {
                "padding_type": "zeros",
                "unsharp_coeff": 5.36,
                "unsharp_sigma": 4.1,
            },
        }
        self._window.setSlicesRange(0, 20)
        configuration = {
            "slice_index": {"Slice": 2},
            "delta_beta_values": (50, 100, 200),
            "score_method": "standard deviation",
            "nabu_params": nabu_specific_params,
        }
        self._window.setConfiguration(configuration)
        res_conf = self._window.getConfiguration()
        for key in (
            "slice_index",
            "delta_beta_values",
            "score_method",
        ):
            self.assertEqual(configuration[key], res_conf[key])
        for key, value in nabu_specific_params["phase"].items():
            assert value == res_conf["nabu_params"]["phase"][key]

    def testSetResults(self):
        """Test setting results"""
        self._window.setDBScores(self._cor_scores, score_method="total variation")
        # test saving snapshots
        with tempfile.TemporaryDirectory() as output_png_imgs:
            final_dir = os.path.join(output_png_imgs, "test/create/it")
            self._window.saveReconstructedSlicesTo(final_dir)
            assert os.path.exists(final_dir)
            assert len(os.listdir(final_dir)) == len(self._cor_scores)



from contextlib import contextmanager

from silx.gui import qt

if qt.BINDING == "PyQt5":
    from PyQt5.QtTest import QSignalSpy  # pylint: disable=E0401,E0611
elif qt.BINDING == "PySide6":
    from PySide6.QtTest import QSignalSpy  # pylint: disable=E0401,E0611
elif qt.BINDING == "PyQt6":
    from PyQt6.QtTest import QSignalSpy  # pylint: disable=E0401,E0611
else:
    QSignalSpy = None


@contextmanager
def block_signals(w: qt.QWidget):
    old = w.blockSignals(True)
    try:
        yield
    finally:
        w.blockSignals(old)



from __future__ import annotations

import gc
import logging
import h5py
import numpy

from collections import namedtuple
from typing import Any

from silx.io.utils import h5py_read_dataset

from tomoscan.io import HDF5File

from nxtomo.paths.nxtomo import get_paths as get_nexus_paths
from nxtomo.nxobject.nxtransformations import NXtransformations
from nxtomo.utils.transformation import (
    DetYFlipTransformation,
    DetZFlipTransformation,
)

from tomwer.core.process.task import Task
from tomwer.core.scan.nxtomoscan import NXtomoScan

_EditorFieldInfo = namedtuple(
    "_EditorFieldInfo", ["nexus_path", "name", "expected_type", "units", "n_value"]
)

_logger = logging.getLogger(__name__)


class NXtomoEditorKeys:
    """namespace to store all keys used by the nxtomo editor"""

    ENERGY = "instrument.beam.energy"
    X_PIXEL_SIZE = "instrument.detector.x_pixel_size"
    Y_PIXEL_SIZE = "instrument.detector.y_pixel_size"
    SAMPLE_DETECTOR_DISTANCE = "instrument.detector.distance"
    FIELD_OF_VIEW = "instrument.detector.field_of_view"
    X_FLIPPED = "instrument.detector.x_flipped"
    Y_FLIPPED = "instrument.detector.y_flipped"
    X_TRANSLATION = "sample.x_translation"
    Y_TRANSLATION = "sample.y_translation"
    Z_TRANSLATION = "sample.z_translation"


class NXtomoEditorTask(
    Task,
    input_names=("data", "configuration"),
    output_names=("data",),
):
    """
    task to edit a couple of field of a NXtomo
    """

    Y_FLIP_MATRIX = DetYFlipTransformation(flip=True).as_matrix()
    # up-down flip matrix
    Z_FLIP_MATRIX = DetZFlipTransformation(flip=True).as_matrix()
    # left-right flip matrix
    Y_AND_Z_flip_MATRIX = numpy.matmul(
        Y_FLIP_MATRIX,
        Z_FLIP_MATRIX,
    )

    def run(self):
        scan = self.inputs.data
        if scan is None:
            _logger.warning("no scan found to be saved")
            return
        if not isinstance(scan, NXtomoScan):
            raise TypeError(
                f"data is expected to be an instance of {NXtomoScan}. Got {type(scan)}"
            )

        configuration = self.inputs.configuration
        if not isinstance(configuration, dict):
            raise TypeError(
                f"configuration is expected to be an instance of {dict}. Got {type(configuration)}"
            )

        n_frames = len(scan.image_key_control)
        mapping = self.build_mapping(
            nexus_version=scan.nexus_version, n_frames=n_frames
        )

        x_flip = configuration.pop(NXtomoEditorKeys.X_FLIPPED, None)
        y_flip = configuration.pop(NXtomoEditorKeys.Y_FLIPPED, None)

        with HDF5File(scan.master_file, mode="a") as h5f:
            entry = h5f[scan.entry]

            for field, field_value in configuration.items():
                if field not in mapping:
                    raise ValueError(f"field unknown ({field})")
                self.__write_to_file(
                    entry=entry,
                    path=mapping[field].nexus_path,
                    value=field_value,
                    name=mapping[field].name,
                    expected_type=mapping[field].expected_type,
                    units=mapping[field].units,
                    n_value=mapping[field].n_value,
                )

            # solve NXtransformations
            if x_flip is not None or y_flip is not None:
                nx_transformations, detector_transformation_path = (
                    self.get_detector_transformations(
                        scan_entry=scan.entry,
                        x_flip=x_flip if x_flip is not None else scan.x_flipped,
                        y_flip=y_flip if y_flip is not None else scan.y_flipped,
                        nexus_version=scan.nexus_version,
                    )
                )
            else:
                nx_transformations = detector_transformation_path = None

        # make sure the file has been removed (fix https://gitlab.esrf.fr/tomotools/tomwer/-/merge_requests/874#note_366688).
        gc.collect()
        if nx_transformations is not None:
            nx_transformations.save(
                file_path=scan.master_file,
                data_path=detector_transformation_path,
                nexus_path_version=scan.nexus_version,
                overwrite=True,
            )
        # clear caches to make sure all modifications will be considered
        scan.clear_caches()
        scan.clear_frames_caches()
        self.outputs.data = scan

    @staticmethod
    def get_detector_transformations(
        scan_entry: str, x_flip: bool, y_flip: bool, nexus_version: float | None
    ) -> dict:
        nexus_paths = get_nexus_paths(nexus_version)
        if nexus_paths.nx_detector_paths.NX_TRANSFORMATIONS is None:
            # old NXtomo are not handling NX_TRANSFORMATIONS
            _logger.debug(
                "Old version of NXtomo found. No information about transformation will be saved"
            )
            return

        nx_transformations = NXtransformations()
        nx_transformations.add_transformation(DetYFlipTransformation(flip=y_flip))
        nx_transformations.add_transformation(DetZFlipTransformation(flip=x_flip))

        detector_transformation_path = "/".join(
            (
                scan_entry,
                nexus_paths.INSTRUMENT_PATH,
                nexus_paths.nx_instrument_paths.DETECTOR_PATH,
                nexus_paths.nx_detector_paths.NX_TRANSFORMATIONS,
            ),
        )
        return nx_transformations, detector_transformation_path

    @staticmethod
    def __write_to_file(
        entry: h5py.Group,
        path: str,
        value: Any,
        name: str,
        expected_type: type,
        n_value: int = 1,
        units=None,
    ) -> None:
        if path is None:
            # if the path does not exists (no handled by this version of nexus for example)
            return

        # try to cast the value
        if isinstance(value, str):
            value = value.replace(" ", "")
            if value.lower() == "none" or "..." in value:
                # if value is not defined or is an array not overwrite by the user (case of the ... )
                return
        elif value is None:
            pass
        else:
            try:
                value = expected_type(value)
            except (ValueError, TypeError) as e:
                _logger.error(f"Fail to overwrite {name} of {entry.name}. Error is {e}")
                return

        if path in entry:
            if not NXtomoEditorTask.isFieldValueDifferent(
                dataset=entry[path], new_value=value, units=units
            ):
                # if no need to overwrite
                return
            else:
                del entry[path]
        if value is None:
            return
        elif n_value == 1:
            entry[path] = value
        else:
            entry[path] = numpy.array([value] * n_value)
        if units is not None:
            entry[path].attrs["units"] = units

    @staticmethod
    def isFieldValueDifferent(dataset: h5py.Dataset, new_value, units) -> bool:
        """
        return False if the given value is the same as the one stored.

        This is a small improvement to avoid rewrite field if we can avoid it.
        The reason behind is that the orange widget will update the field when it receive a scan (of the unlock field).
        But then users can edit any field. As we don't want to bother with complex stuff like for each field keep track if
        It has been modified or not the orange widget will always require to modify all possible field even is not modified.
        """
        current_value = h5py_read_dataset(dataset)
        attrs = dataset.attrs
        current_unit = attrs.get("units", attrs.get("unit", None))
        if units != current_unit:
            # if the unit is not the same, even if the value is the same we will overwrite it
            return True
        else:
            if isinstance(new_value, numpy.ndarray) and isinstance(
                current_value, numpy.ndarray
            ):
                return not numpy.array_equal(new_value, current_value)
            elif numpy.isscalar(current_value) and numpy.isscalar(new_value):
                return current_value != new_value
            else:
                return True

    @staticmethod
    def build_mapping(
        nexus_version: float | None, n_frames: int
    ) -> dict[NXtomoEditorKeys, _EditorFieldInfo]:
        """
        for the different field that can be edited we provide an instance of _EditorFieldInfo
        with all metadata needed to save this particular field.

        Note: X_flip and y_flip are saved differently as the two parameters are coupled we want either to save both or none of it.
        """
        nexus_paths = get_nexus_paths(nexus_version)

        return {
            # energy
            NXtomoEditorKeys.ENERGY: _EditorFieldInfo(
                nexus_path=nexus_paths.ENERGY_PATH,
                expected_type=float,
                units="kev",
                name="energy",
                n_value=1,
            ),
            # x pixel size
            NXtomoEditorKeys.X_PIXEL_SIZE: _EditorFieldInfo(
                nexus_path=nexus_paths.X_PIXEL_SIZE_PATH,
                expected_type=float,
                units="m",
                name="x pixel size",
                n_value=1,
            ),
            # y pixel size
            NXtomoEditorKeys.Y_PIXEL_SIZE: _EditorFieldInfo(
                nexus_path=nexus_paths.Y_PIXEL_SIZE_PATH,
                expected_type=float,
                units="m",
                name="y pixel size",
                n_value=1,
            ),
            # sample / detector distance
            NXtomoEditorKeys.SAMPLE_DETECTOR_DISTANCE: _EditorFieldInfo(
                nexus_path=nexus_paths.DISTANCE_PATH,
                expected_type=float,
                units="m",
                name="sample detector distance",
                n_value=1,
            ),
            # overwrite FOV
            NXtomoEditorKeys.FIELD_OF_VIEW: _EditorFieldInfo(
                nexus_path=nexus_paths.FOV_PATH,
                expected_type=str,
                units=None,
                name="field of view",
                n_value=1,
            ),
            # x translation
            NXtomoEditorKeys.X_TRANSLATION: _EditorFieldInfo(
                nexus_path=nexus_paths.X_TRANS_PATH,
                expected_type=float,
                units="m",
                name="x translation",
                n_value=n_frames,
            ),
            # y translation
            NXtomoEditorKeys.Y_TRANSLATION: _EditorFieldInfo(
                nexus_path=nexus_paths.Y_TRANS_PATH,
                expected_type=float,
                units="m",
                name="y translation",
                n_value=n_frames,
            ),
            # z translation
            NXtomoEditorKeys.Z_TRANSLATION: _EditorFieldInfo(
                nexus_path=nexus_paths.Z_TRANS_PATH,
                expected_type=float,
                units="m",
                name="z translation",
                n_value=n_frames,
            ),
        }



# coding: utf-8
from __future__ import annotations

import os
import shutil
import tempfile

from silx.gui import qt
from silx.gui.utils.testutils import TestCaseQt

from tomwer.core.process.reconstruction.normalization.params import (
    Method as NormalizationMethod,
)
from tomwer.core.process.reconstruction.normalization.params import (
    _ValueSource as NormalizationSource,
)
from tomwer.core.utils.scanutils import MockNXtomo
from tomwer.gui.reconstruction.normalization.intensity import SinoNormWindow


class TestNormIntensityWindow(TestCaseQt):
    def setUp(self):
        super(TestNormIntensityWindow, self).setUp()
        self._widget = SinoNormWindow(parent=None)
        self._tmp_dir = tempfile.mkdtemp()
        scan_path = os.path.join(self._tmp_dir, "my_scan_2")
        self.scan = MockNXtomo(
            scan_path=scan_path,
            n_ini_proj=10,
            n_proj=10,
            n_alignement_proj=2,
            create_final_flat=False,
            create_ini_dark=True,
            create_ini_flat=True,
            n_refs=1,
        ).scan

    def tearDown(self):
        self._widget.setAttribute(qt.Qt.WA_DeleteOnClose)
        self._widget.close()
        self.qapp.processEvents()
        self._widget = None
        shutil.rmtree(self._tmp_dir)
        super(TestNormIntensityWindow, self).tearDown()

    def test(self):
        self._widget.setScan(self.scan)
        self._widget.show()
        self.qWaitForWindowExposed(self._widget)
        self._widget.setCurrentMethod(NormalizationMethod.NONE)
        self.qapp.processEvents()
        assert self._widget.getConfiguration() == {"source": "none", "method": "none"}

        self._widget.setCurrentMethod(NormalizationMethod.CHEBYSHEV)
        self.qapp.processEvents()
        assert self._widget.getConfiguration() == {
            "method": "chebyshev",
            "source": "none",
        }

        self._widget.setCurrentMethod(NormalizationMethod.SUBTRACTION)
        self.qapp.processEvents()
        self._widget.setCurrentSource(NormalizationSource.MANUAL_ROI)
        self.qapp.processEvents()
        output_configuration = self._widget.getConfiguration()
        assert output_configuration["method"] == "subtraction"
        assert output_configuration["source"] == "manual ROI"
        assert "start_x" in output_configuration
        assert "end_x" in output_configuration
        assert "start_y" in output_configuration
        assert "end_y" in output_configuration
        assert "calc_fct" in output_configuration
        self._widget.setCurrentMethod(NormalizationMethod.DIVISION)
        self.qapp.processEvents()
        self._widget.setCurrentSource(NormalizationSource.DATASET)
        self.qapp.processEvents()
        output_configuration = self._widget.getConfiguration()
        assert output_configuration["method"] == "division"
        assert output_configuration["source"] == "from dataset"
        assert "dataset_url" in output_configuration

        self._widget.setCurrentMethod(NormalizationMethod.DIVISION)
        self.qapp.processEvents()
        self._widget.setCurrentSource(NormalizationSource.MANUAL_SCALAR)
        self.qapp.processEvents()
        output_configuration = self._widget.getConfiguration()
        assert output_configuration["method"] == "division"
        assert output_configuration["source"] == "scalar"
        assert "value" in output_configuration



# coding: utf-8
from __future__ import annotations

import logging

import fabio

_logger = logging.getLogger(__name__)


def flatFieldCorrection(imgs, dark, flat):
    """
    Simple normalization of a list of images.
    Normalization is made for X-Ray imaging:
    (img - dark) / (flat - dark)

    :param imgs: list of imgs to correct. key: index of the image,
                      value: the image path or numpy.ndarray
    :param dark: dark image
    :param flat: flat image
    :return: list of corrected images
    """
    res = {}
    conditionOK = True
    if dark.ndim != 2:
        _logger.error(
            "cannot make flat field correction, dark should be of " "dimension 2"
        )
        conditionOK = False

    if flat.ndim != 2:
        _logger.error(
            "cannot make flat field correction, flat should be of " "dimension 2"
        )
        conditionOK = False

    if dark.shape != flat.shape:
        _logger.error("Given dark and flat have incoherent dimension")
        conditionOK = False

    if conditionOK is False:
        return res

    for index, img in imgs.items():
        imgData = img
        if type(img) is str:
            assert img.endswith(".edf")
            imgData = fabio.open(img).data

        if imgData.shape != dark.shape:
            _logger.error("Image has invalid. Cannot apply flat field" "correction it")
            corrrectedImage = imgData
        else:
            corrrectedImage = (imgData - dark) / (flat - dark)

        res[index] = corrrectedImage

    return res



# coding: utf-8
from __future__ import annotations

import logging
import os
import shutil
import tempfile
import unittest

from silx.gui import qt
from silx.gui.utils.testutils import TestCaseQt

from tomwer.core.utils.scanutils import MockNXtomo
from tomwer.gui.visualization.diffviewer import DiffFrameViewer

logging.disable(logging.INFO)


class TestDiffViewer(TestCaseQt):
    """unit test for the :class:_ImageStack widget"""

    def setUp(self):
        TestCaseQt.setUp(self)
        self._widget = DiffFrameViewer()
        self._widget.setAttribute(qt.Qt.WA_DeleteOnClose)

        self.tmp_dir = tempfile.mkdtemp()
        self.scan1 = MockNXtomo(
            scan_path=os.path.join(self.tmp_dir, "myscan"),
            n_proj=20,
            n_ini_proj=20,
            dim=10,
        ).scan

        self.scan2 = MockNXtomo(
            scan_path=os.path.join(self.tmp_dir, "myscan"),
            n_proj=20,
            n_ini_proj=20,
            dim=10,
        ).scan

    def tearDown(self):
        shutil.rmtree(self.tmp_dir)
        self._widget.close()
        self._widget = None
        unittest.TestCase.tearDown(self)

    def test(self):
        """Make sur the addLeaf and clear functions are working"""
        self._widget.addScan(self.scan1)
        self._widget.addScan(self.scan2)
        self._widget.getLeftScan()

        # test shift
        shift_widgets = self._widget.getShiftsWidget()
        relShiftWidget = shift_widgets.getRelativeShiftWidget()
        relShiftWidget.setShiftStep(0.1)
        relShiftWidget.move("left")
        self.qapp.processEvents()
        relShiftWidget.setShiftStep(0.3)
        relShiftWidget.move("right")
        self.qapp.processEvents()
        relShiftWidget.setShiftStep(10)
        relShiftWidget.move("up")
        self.qapp.processEvents()
        relShiftWidget.setShiftStep(33)
        relShiftWidget.move("down")
        self.qapp.processEvents()
        # for the frame A y is expected to be always 0
        assert relShiftWidget.getFrameAShift() == (0.2, 0)
        # for the frame B y is expected to be the shift. And x is supposed to be the opposite of the A.x shift
        assert relShiftWidget.getFrameBShift() == (-0.2, -23)



from __future__ import annotations

from orangewidget import settings


class CallbackSettingsHandler(settings.SettingsHandler):
    """
    Settings handler used to call some callback before packing data (so before)
    saving orange :class:`Setting`
    """

    def __init__(self):
        super(CallbackSettingsHandler, self).__init__()
        self.__callbacks = []

    def addCallback(self, _callback):
        self.__callbacks.append(_callback)

    def removeCallback(self, callback):
        self.__callbacks.remove(callback)

    def pack_data(self, widget):
        """"""
        for callback in self.__callbacks:
            callback()
        return super(CallbackSettingsHandler, self).pack_data(widget)



# coding: utf-8

from tomwer.core.process.visualization.volumeviewer import _VolumeViewerPlaceHolder


def test_volume_viewer():
    process = _VolumeViewerPlaceHolder(
        inputs={
            "data": None,
        }
    )
    process.run()



# coding: utf-8

from tomwer.core.process.visualization.radiostack import _RadioStackPlaceHolder


def test_radio_stack():
    process = _RadioStackPlaceHolder(
        inputs={
            "data": None,
        }
    )
    process.run()



# coding: utf-8
from __future__ import annotations


from silx.gui import qt

from tomwer.core.process.reconstruction.nabu.utils import _NabuStages
from tomwer.gui.reconstruction.nabu.nabuconfig.output import _NabuOutputConfig
from tomwer.gui.reconstruction.nabu.nabuconfig.phase import _NabuPhaseConfig
from tomwer.gui.reconstruction.nabu.nabuconfig.preprocessing import (
    _NabuPreProcessingConfig,
)
from tomwer.gui.reconstruction.nabu.nabuconfig.reconstruction import (
    _NabuReconstructionConfig,
)


class NabuConfiguration(qt.QWidget):
    """
    Top level widget for defining the nabu configuration
    """

    sigConfChanged = qt.Signal(str, str)
    """Signal emitted when the configuration change. Parameters are
    (stage, index option modified)
    """

    def __init__(self, parent):
        qt.QWidget.__init__(self, parent=parent)
        self._stageHided = []
        # list of hided stage. Not taking into account the filter

        self.setLayout(qt.QVBoxLayout())

        # pre processing options
        self._preProcessingGB = qt.QGroupBox("pre processing", self)
        self._preProcessingGB.setLayout(qt.QVBoxLayout())

        self._preProcessingWidget = _NabuPreProcessingConfig(parent=self)
        self._preProcessingGB.layout().addWidget(self._preProcessingWidget)
        self.layout().addWidget(self._preProcessingGB)

        # phase options
        self._phaseGB = qt.QGroupBox("phase", self)
        self._phaseGB.setLayout(qt.QVBoxLayout())

        self._phaseWidget = _NabuPhaseConfig(parent=self)
        self._phaseGB.layout().addWidget(self._phaseWidget)
        self.layout().addWidget(self._phaseGB)

        # reconstruction opts
        self._reconstructionGB = qt.QGroupBox("reconstruction", self)
        self._reconstructionGB.setLayout(qt.QVBoxLayout())

        self._reconstructionWidget = _NabuReconstructionConfig(parent=self)
        self._reconstructionGB.layout().addWidget(self._reconstructionWidget)
        self.layout().addWidget(self._reconstructionGB)

        # output information
        self._outputGB = qt.QGroupBox("output", self)
        self._outputGB.setLayout(qt.QVBoxLayout())

        self._outputWidget = _NabuOutputConfig(parent=self)
        self._outputGB.layout().addWidget(self._outputWidget)
        self.layout().addWidget(self._outputGB)

        # connect signal / slot
        self._preProcessingWidget.sigConfChanged.connect(
            self._signalConfChangedPreProcessing
        )
        self._phaseWidget.sigConfChanged.connect(self._signalConfChangedPhase)
        self._reconstructionWidget.sigConfChanged.connect(
            self._signalConfChangedReconstruction
        )
        self._outputWidget.sigConfChanged.connect(self._signalConfChangedOutput)

        # expose API
        self.getSlices = self._reconstructionWidget.getSlices
        self.setOutputDir = self._outputWidget.setOutputDir

    def setDeltaBetaValue(self, value):
        self._phaseWidget.setDeltaBetaValue(value)

    def hideSlicesInterface(self):
        self._reconstructionWidget.hideSlicesInterface()

    def hidePaganinInterface(self):
        self._stageHided.append(self._phaseGB)
        self._phaseGB.hide()

    def _signalConfChangedPhase(self, param):
        self.sigConfChanged.emit(self._phaseWidget.getStage().value, param)

    def _signalConfChangedPreProcessing(self, param):
        self.sigConfChanged.emit(self._preProcessingWidget.getStage().value, param)

    def _signalConfChangedReconstruction(self, param):
        self.sigConfChanged.emit(self._reconstructionWidget.getStage().value, param)

    def _signalConfChangedOutput(self, param):
        self.sigConfChanged.emit(self._outputWidget.getStage().value, param)

    def getConfiguration(self):
        config = {
            "preproc": self._preProcessingWidget.getConfiguration(),
            "reconstruction": self._reconstructionWidget.getConfiguration(),
            "dataset": self._reconstructionWidget.getDatasetConfiguration(),
            "tomwer_slices": self._reconstructionWidget.getSlices(),
            "output": self._outputWidget.getConfiguration(),
            "phase": self._phaseWidget.getConfiguration(),
        }
        return config

    def setConfiguration(self, config):
        if "preproc" in config:
            self._preProcessingWidget.setConfiguration(config["preproc"])
        if "phase" in config:
            self._phaseWidget.setConfiguration(config["phase"])
        if "reconstruction" in config:
            self._reconstructionWidget.setConfiguration(config["reconstruction"])
        if "tomwer_slices" in config:
            self._reconstructionWidget.setSlices(config["tomwer_slices"])
        if "dataset" in config:
            self._reconstructionWidget.setDatasetConfiguration(config["dataset"])
        if "output" in config:
            self._outputWidget.setConfiguration(config["output"])

    def applyFilter(self, stage, option):
        if stage is None:
            for widget in (
                self._preProcessingGB,
                self._reconstructionGB,
                self._outputGB,
                self._phaseGB,
            ):
                widget.setVisible(True)
        else:
            stage = _NabuStages.from_value(stage)
            self._preProcessingGB.setVisible(stage is _NabuStages.PRE)
            self._reconstructionGB.setVisible(stage is _NabuStages.PROC)
            self._outputGB.setVisible(stage is _NabuStages.POST)
            self._phaseGB.setVisible(stage is _NabuStages.PHASE)

        # handle _stageHided list
        for widget in (
            self._preProcessingGB,
            self._reconstructionGB,
            self._outputGB,
            self._phaseGB,
        ):
            if widget in (self._stageHided):
                self._phaseGB.setVisible(False)

    def setConfigurationLevel(self, level):
        for widget in (
            self._preProcessingWidget,
            self._reconstructionWidget,
            self._outputWidget,
            self._phaseWidget,
        ):
            widget.setConfigurationLevel(level)


class NabuConfigurationTab(qt.QTabWidget):
    """
    Top level widget for defining the nabu configuration.
    Same as NabuConfiguration but inside a tab
    """

    sigConfChanged = qt.Signal(str, str)
    """Signal emitted when the configuration change. Parameters are
    (stage, index option modified)
    """

    def __init__(self, parent):
        qt.QTabWidget.__init__(self, parent=parent)

        # pre processing options
        self._preProcessingWidget = _NabuPreProcessingConfig(parent=self)
        self.addTab(self._preProcessingWidget, "pre processing")

        # phase options
        self._phaseGB = qt.QGroupBox("apply phase", self)
        self._phaseGB.setLayout(qt.QVBoxLayout())

        self._phaseWidget = _NabuPhaseConfig(parent=self)
        self._phaseGB.layout().addWidget(self._phaseWidget)
        self.addTab(self._phaseGB, "phase")

        # reconstruction opts
        self._reconstructionWidget = _NabuReconstructionConfig(parent=self)
        self.addTab(self._reconstructionWidget, "reconstruction")

        # output information
        self._outputWidget = _NabuOutputConfig(parent=self)
        self.addTab(self._outputWidget, "output")

        # connect signal / slot
        self._preProcessingWidget.sigConfChanged.connect(
            self._signalConfChangedPreProcessing
        )
        self._phaseWidget.sigConfChanged.connect(self._signalConfChangedPhase)
        self._reconstructionWidget.sigConfChanged.connect(
            self._signalConfChangedReconstruction
        )

        # expose API
        self.getSlices = self._reconstructionWidget.getSlices
        self.setOutputDir = self._outputWidget.setOutputDir

    def _signalConfChangedPhase(self, param):
        self.sigConfChanged.emit(self._phaseWidget.getStage().value, param)

    def _signalConfChangedPreProcessing(self, param):
        self.sigConfChanged.emit(self._preProcessingWidget.getStage().value, param)

    def _signalConfChangedReconstruction(self, param):
        self.sigConfChanged.emit(self._reconstructionWidget.getStage().value, param)

    def getConfiguration(self):
        config = {
            "preproc": self._preProcessingWidget.getConfiguration(),
            "reconstruction": self._reconstructionWidget.getConfiguration(),
            "tomwer_slices": self.getSlices(),
            "output": self._outputWidget.getConfiguration(),
        }
        config["phase"] = self._phaseWidget.getConfiguration()
        return config

    def setConfiguration(self, config):
        if "preproc" in config:
            self._preProcessingWidget.setConfiguration(config["preproc"])
        if "phase" in config:
            self._phaseWidget.setConfiguration(config["phase"])
        if "reconstruction" in config:
            self._reconstructionWidget.setConfiguration(config["reconstruction"])
        if "tomwer_slices" in config:
            self._reconstructionWidget.setSlices(config["tomwer_slices"])



from __future__ import annotations

from silx.gui import qt
from nabu.stitching.config import dict_to_config_obj
from nabu.stitching.stitcher.z_stitcher import (
    PreProcessingZStitcher,
    PostProcessingZStitcher,
)
from nabu.stitching.stitcher.y_stitcher import PreProcessingYStitcher


class PreviewThread(qt.QThread):
    """
    Thread to compute an overview of the stitching
    """

    def __init__(self, stitching_config: dict, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._stitching_config = dict_to_config_obj(stitching_config)
        self._output_identifier = None
        self._frame_composition = None
        self._final_tomo_objs_positions = None
        # store position of all the tomo objects (scan, volumes) used for the final stitching (after shift refinement)

    @property
    def stitching_config(self):
        return self._stitching_config

    @property
    def output_identifier(self):
        return self._output_identifier

    @property
    def frame_composition(self):
        return self._frame_composition

    @property
    def final_tomo_objs_positions(self) -> dict:
        """
        :return: dict with tomo object identifier (str) as key and a tuple of position in pixel (axis_0_pos, axis_1_pos, axis_2_pos)
        """
        return self._final_tomo_objs_positions

    def run(self):
        stitching_type = self.stitching_config.stitching_type
        if stitching_type.value == "z-preproc":
            stitcher = PreProcessingZStitcher(configuration=self.stitching_config)
        elif stitching_type.value == "z-postproc":
            stitcher = PostProcessingZStitcher(configuration=self.stitching_config)
        elif stitching_type.value == "y-preproc":
            stitcher = PreProcessingYStitcher(configuration=self.stitching_config)
        else:
            raise NotImplementedError
        self._output_identifier = stitcher.stitch()
        if self._output_identifier is not None:
            self._output_identifier = self._output_identifier.to_str()
        # store in cache the frame composition to be able to provide them to the PreviewPlot
        self._frame_composition = stitcher.frame_composition
        self._final_tomo_objs_positions = stitcher.get_final_axis_positions_in_px()



# coding: utf-8
from __future__ import annotations

import gc
import os
import pickle
import tempfile
import time

import numpy
from orangecanvas.scheme.readwrite import literal_dumps
from silx.gui import qt
from silx.gui.utils.testutils import TestCaseQt

from orangecontrib.tomwer.widgets.reconstruction.CastNabuVolumeOW import (
    CastNabuVolumeOW,
)
from tomwer.core.volume.rawvolume import RawVolume


class TestCastVolumeOW(TestCaseQt):
    def setUp(self):
        super().setUp()
        self._window = CastNabuVolumeOW()

    def tearDown(self):
        self._window.setAttribute(qt.Qt.WA_DeleteOnClose)
        self._window.close()
        self._window = None
        gc.collect()

    def test(self):
        self._window.show()
        self.qWaitForWindowExposed(self._window)

    def test_serializing(self):
        pickle.dumps(self._window.getConfiguration())

    def test_literal_dumps(self):
        literal_dumps(self._window.getConfiguration())

    def test_cast_volume(self):
        with tempfile.TemporaryDirectory() as tmp_dir:
            input_dir = os.path.join(tmp_dir, "input")
            os.makedirs(input_dir)

            volume = RawVolume(
                file_path=os.path.join(input_dir, "vol_file.vol"),
                data=numpy.linspace(0, 10, 100 * 100 * 3, dtype=numpy.float32).reshape(
                    3, 100, 100
                ),
            )
            volume.save()

            self._window.process_volume(volume)
            while self._window._processingStack.is_computing():
                time.sleep(0.1)
                self.qapp.processEvents()



from __future__ import annotations

from orangewidget import gui, settings, widget
from orangewidget.widget import Output
from silx.gui import qt

import tomwer.core.process.control.datalistener.datalistener
from orangecontrib.tomwer.widgets.utils import WidgetLongProcessing
from tomwer.core.process.control.datalistener import DataListener
from tomwer.core.process.control.datalistener.rpcserver import (
    send_signal_to_local_rpc_servers,
)
from tomwer.core.process.control.nxtomomill import H5ToNxProcess
from tomwer.core.scan.blissscan import BlissScan
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.gui.control.datalistener import DataListenerWidget
from tomwer.synctools.datalistener import DataListenerQThread, MockDataListenerQThread
from tomwer.synctools.stacks.control.datalistener import DataListenerProcessStack
from tomwer.utils import docstring

try:
    # new HDF5Config class
    from nxtomomill.io.config import TomoHDF5Config as HDF5Config
except ImportError:
    from nxtomomill.io.config import HDF5Config

import functools
import logging
import signal
import sys
from typing import Iterable

logger = logging.getLogger(__name__)


class DataListenerOW(
    widget.OWBaseWidget, WidgetLongProcessing, DataListener, openclass=True
):
    """
    This widget is used to listen to a server notifying the widget when an
    acquisition is finished.
    Then the bliss file will be converted to .nx file, NXtomo compliant.
    """

    name = "scan listener"
    id = "orangecontrib.widgets.tomwer.control.DataListenerOW.DataListenerOW"
    description = (
        "The widget will receive information from bliss acquisition "
        "and wait for acquisition to be finished. Once finished it "
        "will call nxtomomill to convert from bliss .hdf5 to "
        "NXtomo compliant .nx file"
    )
    icon = "icons/datalistener.svg"
    priority = 10
    keywords = [
        "tomography",
        "file",
        "tomwer",
        "listener",
        "datalistener",
        "hdf5",
        "NXtomo",
    ]

    want_main_area = True
    want_control_area = False
    resizing_enabled = True

    ewokstaskclass = (
        tomwer.core.process.control.datalistener.datalistener._DataListenerTaskPlaceHolder
    )

    _blissConfiguration = settings.Setting(dict())
    # to keep backward compatibility

    _nxtomo_cfg_file = settings.Setting(str())
    # to keep backward compatibility

    _static_input = settings.Setting(dict())

    class Outputs:
        data = Output(name="data", type=TomwerScanBase, doc="one scan to be process")

    def __init__(self, parent=None):
        widget.OWBaseWidget.__init__(self, parent)
        WidgetLongProcessing.__init__(self)
        DataListener.__init__(self)
        self._processingStack = DataListenerProcessStack()
        self._processingStack.sigComputationEnded.connect(self._signal_scan_ready)
        self._widget = DataListenerWidget(parent=self)
        self._mock = False

        self._box = gui.vBox(self.mainArea, self.name)
        layout = self._box.layout()
        layout.addWidget(self._widget)

        # signal / slot connection
        self._widget.sigActivate.connect(self._activated)
        self._widget.sigDeactivate.connect(self._deactivated)
        self._widget.sigConfigurationChanged.connect(self._jsonRPCConfigChanged)
        self._widget.sigCFGFileChanged.connect(self._nxtomoFileChanged)
        self._widget.sigAcquisitionEnded.connect(self._process_bliss_file_frm_tuple)
        self._widget.sigServerStopped.connect(self._serverStopped)

        # manage server stop when delete directly the widget or stop by Ctr+C
        signal.signal(signal.SIGINT, self.handleSigTerm)
        onDestroy = functools.partial(self._stopServerBeforeclosing)
        self.destroyed.connect(onDestroy)

        # set up
        self._loadSettings()

        # for conveniance start the listener when create it.
        self.activate(True)

    def _loadSettings(self):
        if "bliss_server_configuration" in self._static_input:  # pylint: disable=E1135
            bliss_configuration = self._static_input[  # pylint: disable=E1136
                "bliss_server_configuration"
            ]
        else:
            bliss_configuration = self._blissConfiguration
        if bliss_configuration != {}:
            self._widget.setBlissServerConfiguation(bliss_configuration)
        if "nxtomomill_cfg_file" in self._static_input:  # pylint: disable=E1135
            nxtomo_cfg_file = self._static_input[  # pylint: disable=E1136
                "nxtomomill_cfg_file"
            ]
        else:
            nxtomo_cfg_file = self._nxtomo_cfg_file
        self._widget.setCFGFilePath(nxtomo_cfg_file)
        if "output_dir" in self._static_input:  # pylint: disable=E1135
            self._widget.setOutputFolder(
                self._static_input["output_dir"]  # pylint: disable=E1136
            )

    def getNXTomomillConfiguration(self):
        cfg_file = self._widget.getCFGFilePath()

        def create_default_config():
            configuration = HDF5Config()
            # insure minimal bacward compatibility
            if hasattr(configuration, "bam_single_file"):
                configuration.bam_single_file = True
            return configuration

        if cfg_file in (None, ""):
            config = create_default_config()
        else:
            try:
                config = HDF5Config.from_cfg_file(cfg_file)
            except Exception as e:
                logger.warning(f"Fail to load configuraiton file. Error is {e}")
                config = create_default_config()

        return config

    def _process_bliss_file_frm_tuple(self, t):
        master_file, entry, proposal_file, saving_file, success = t
        bliss_scan = BlissScan(
            master_file=master_file,
            entry=str(entry) + ".1",
            proposal_file=proposal_file,
            saving_file=saving_file,
        )
        configuration = self.getNXTomomillConfiguration()
        # overwrite output file
        configuration.output_file = H5ToNxProcess.deduce_output_file_path(
            bliss_scan.master_file,
            entry=bliss_scan.entry,
            outputdir=self._widget.getOutputFolder(),
            scan=bliss_scan,
        )
        if success:
            self._processingStack.add(data=bliss_scan, configuration=configuration)
        else:
            # TODO: maybe remove the acquisition ?
            pass

    def _activated(self):
        self.activate(True)

    def _deactivated(self):
        self.activate(False)

    def _serverStopped(self):
        """
        Callback when the server is stopped
        """
        self.activate(False)

    @docstring(DataListener.activate)
    def activate(self, activate=True):
        if activate and not self.is_port_available():
            old = self._widget.blockSignals(True)
            self._widget.activate(activate=False)
            self._widget.blockSignals(old)
            dialog = _PortOccupyDialog(parent=self, port=self.port, host=self.host)
            dialog.setModal(False)
            if dialog.exec_() == qt.QDialog.Accepted:
                if dialog.retry_connection:
                    return self.activate(activate=True)
            else:
                return

        old = self._widget.blockSignals(True)
        self.set_configuration(self._widget.getBlissServerConfiguration())
        self._widget.activate(activate=activate)
        DataListener.activate(self, activate=activate)
        self.processing_state(activate, info="listener active")
        self._widget.blockSignals(old)

    def _signal_scan_ready(self, scan, future_tomo_obj):
        if scan is None:
            return
        assert isinstance(scan, Iterable)
        for s in scan:
            assert isinstance(s, TomwerScanBase)
            self.Outputs.data.send(s)

    def _ask_user_for_overwritting(self, file_path):
        msg = qt.QMessageBox(self)
        msg.setIcon(qt.QMessageBox.Question)
        types = qt.QMessageBox.Ok | qt.QMessageBox.Cancel
        msg.setStandardButtons(types)

        text = "NXtomomill will overwrite: \n %s. Do you agree ?" % file_path
        msg.setText(text)
        return msg.exec_() == qt.QMessageBox.Ok

    def _jsonRPCConfigChanged(self):
        self._blissConfiguration = self._widget.getBlissServerConfiguration()
        self._static_input["bliss_server_configuration"] = (  # pylint: disable=E1137
            self._widget.getBlissServerConfiguration()
        )
        self._static_input["output_dir"] = (  # pylint: disable=E1137
            self._widget.getOutputFolder()
        )
        if self.is_active():
            self.activate(False)
            self.activate(True)

    def _nxtomoFileChanged(self, cfg_file):
        self._nxtomo_cfg_file = cfg_file
        self._static_input["nxtomomill_cfg_file"] = cfg_file  # pylint: disable=E1137

    def setMock(self, mock, acquisitions):
        self._mock = mock
        self._mock_acquisitions = acquisitions

    @docstring(DataListenerWidget.getHost)
    def getHost(self):
        return self._widget.getHost()

    @docstring(DataListenerWidget.getPort)
    def getPort(self):
        return self._widget.getPort()

    def close(self):
        signal.signal(signal.SIGINT, signal.SIG_DFL)
        self.activate(False)
        super().close()

    @docstring(DataListener.create_listening_thread)
    def create_listening_thread(self):
        if self._mock is True:
            thread = MockDataListenerQThread(
                host=self.getHost(),
                port=self.getPort(),
                acquisitions=None,
                mock_acquisitions=self._mock_acquisitions,
            )
        else:
            thread = DataListenerQThread(
                host=self.getHost(), port=self.getPort(), acquisitions=None
            )
        # connect thread
        thread.sigAcquisitionStarted.connect(
            self._widget._acquisitionStarted, qt.Qt.DirectConnection
        )
        thread.sigAcquisitionEnded.connect(
            self._widget._acquisitionEnded, qt.Qt.DirectConnection
        )
        thread.sigScanAdded.connect(
            self._widget._acquisitionUpdated, qt.Qt.DirectConnection
        )
        thread.sigServerStop.connect(
            self._widget._serverStopped, qt.Qt.DirectConnection
        )
        return thread

    @docstring(DataListener.delete_listening_thread)
    def delete_listening_thread(self):
        self._listening_thread.sigAcquisitionStarted.disconnect(
            self._widget._acquisitionStarted
        )
        self._listening_thread.sigAcquisitionEnded.disconnect(
            self._widget._acquisitionEnded
        )
        self._listening_thread.sigScanAdded.disconnect(self._widget._acquisitionUpdated)
        self._listening_thread.sigServerStop.disconnect(self._widget._serverStopped)
        DataListener.delete_listening_thread(self)

    def _stopServerBeforeclosing(self):
        self.activate(False)

    def handleSigTerm(self, signo, *args, **kwargs):
        if signo == signal.SIGINT:
            self._stopServerBeforeclosing()
            sys.exit()

    def _get_n_scan_observe(self):
        return self._widget._observationWidget.observationTable.model().rowCount()

    def _get_n_scan_finished(self):
        return self._widget._historyWindow.scanHistory.model().rowCount()


class _PortOccupyDialog(qt.QDialog):
    def __init__(self, parent, port, host):
        qt.QDialog.__init__(self, parent)
        self.setLayout(qt.QVBoxLayout())
        self._retry = False
        self.port = port
        self.host = host

        mess = (
            f"port ({port}) of {host} already in use. \n Maybe an other "
            "instance of `datalistener` is running in this session or "
            "another tomwer session. \n As this widget is connecting with "
            "bliss we enforce it to be unique."
        )
        self.layout().addWidget(qt.QLabel(mess, self))
        self.setWindowTitle("Unable to launch two listener in parallel")

        types = qt.QDialogButtonBox.Cancel | qt.QDialogButtonBox.Retry
        self._buttons = qt.QDialogButtonBox(self)
        self._buttons.setStandardButtons(types)
        self.layout().addWidget(self._buttons)

        self._sendSIGTERMMsb = qt.QPushButton("send SIGTERM", self)
        self._sendSIGTERMMsb.setToolTip(
            "Try to send SIGTERM signal to the "
            "local tomwer-rpcserver if any "
            "occupies the reserved port"
        )
        self._buttons.addButton(self._sendSIGTERMMsb, qt.QDialogButtonBox.ActionRole)
        self._sendSIGKILLMsb = qt.QPushButton("send SIGKILL", self)
        self._sendSIGKILLMsb.setToolTip(
            "Try to send SIGKILL signal to the "
            "local tomwer-rpcserver if any "
            "occupies the reserved port"
        )
        self._buttons.addButton(self._sendSIGKILLMsb, qt.QDialogButtonBox.ActionRole)

        # set up
        # for now we don't want to show "send signal" feature
        self._sendSIGTERMMsb.hide()
        self._sendSIGKILLMsb.hide()

        # connect signal / slot
        self._buttons.button(qt.QDialogButtonBox.Cancel).released.connect(self.reject)
        self._buttons.button(qt.QDialogButtonBox.Retry).released.connect(
            self._retry_connect
        )
        self._sendSIGTERMMsb.released.connect(self._emitSigterm)
        self._sendSIGKILLMsb.released.connect(self._emitSigkill)

    @property
    def retry_connection(self):
        return self._retry

    def _emitSigterm(self, *args, **kwargs):
        send_signal_to_local_rpc_servers(signal.SIGTERM, port=self.port)

    def _emitSigkill(self, *args, **kwargs):
        send_signal_to_local_rpc_servers(signal.SIGKILL, port=self.port)

    def _retry_connect(self, *args, **kargs):
        self._retry = True
        self.accept()



# coding: utf-8

"""
contains gui for diffviewer shift
"""
from __future__ import annotations


import sys

from silx.gui import qt
from silx.utils.enum import Enum as _Enum

from tomwer.utils import docstring


class _FrameShiftsBase:
    """dummy interface to define how to retrieve shift value"""

    def isFrameALRFlip(self) -> bool:
        """

        :return: True if the frame A should be left-right shift
        """
        raise NotImplementedError

    def getFrameAShift(self) -> tuple:
        """

        :return:
        """
        raise NotImplementedError

    def isFrameBLRFlip(self) -> bool:
        """

        :return: True if the frame B should be left-right shift
        """
        raise NotImplementedError

    def getFrameBShift(self) -> tuple:
        """

        :return:
        """
        raise NotImplementedError


class TwoFramesShiftTab(qt.QTabWidget, _FrameShiftsBase):
    sigShiftsChanged = qt.Signal(tuple)
    """Signal emit when the shift changed. Tuple contains
     shift_imgA, shift_imgB.
     Both are tuple of (x, y, lrflip) x and y are shift values as float.
     lrflip is a boolean notifying if we should flip image or not before applying the shift
    """

    class ShiftMode(_Enum):
        RELATIVE = "relative shift"
        ABSOLUTE = "absolute shift"

    def __init__(self, parent=None):
        super().__init__(parent)
        # create layout
        self._relativeShiftWidget = Relative2FramesShift(self)
        self.addTab(self._relativeShiftWidget, self.ShiftMode.RELATIVE.value)
        self._absoluteShiftWidget = Absolute2FramesShift(self)
        self.addTab(self._absoluteShiftWidget, self.ShiftMode.ABSOLUTE.value)
        # connect signal / slot
        self._absoluteShiftWidget.sigShiftsChanged.connect(self.sigShiftsChanged)
        self._relativeShiftWidget.sigShiftsChanged.connect(self.sigShiftsChanged)
        # set up
        self.setCurrentWidget(self._relativeShiftWidget)
        self._relativeShiftWidget.setFocus(qt.Qt.OtherFocusReason)
        self._relativeShiftWidget.setFocusPolicy(qt.Qt.StrongFocus)
        self.setFocusProxy(self._relativeShiftWidget)

    def getRelativeShiftWidget(self):
        return self._relativeShiftWidget

    def getShiftMode(self):
        if self.currentWidget() is self._relativeShiftWidget:
            return self.ShiftMode.RELATIVE
        elif self.currentWidget() is self._absoluteShiftWidget:
            return self.ShiftMode.ABSOLUTE
        else:
            raise NotImplementedError("mode not handled")

    @docstring(_FrameShiftsBase)
    def isFrameALRFlip(self) -> bool:
        if self.getShiftMode() is self.ShiftMode.ABSOLUTE:
            return self._absoluteShiftWidget.isFrameALRFlip()
        elif self.getShiftMode() is self.ShiftMode.RELATIVE:
            return self._relativeShiftWidget.isFrameALRFlip()
        else:
            raise NotImplementedError("mode not handled")

    @docstring(_FrameShiftsBase)
    def getFrameAShift(self) -> tuple:
        if self.getShiftMode() is self.ShiftMode.ABSOLUTE:
            return self._absoluteShiftWidget.getFrameAShift()
        elif self.getShiftMode() is self.ShiftMode.RELATIVE:
            return self._relativeShiftWidget.getFrameAShift()
        else:
            raise NotImplementedError("mode not handled")

    @docstring(_FrameShiftsBase)
    def isFrameBLRFlip(self) -> bool:
        if self.getShiftMode() is self.ShiftMode.ABSOLUTE:
            return self._absoluteShiftWidget.isFrameBLRFlip()
        elif self.getShiftMode() is self.ShiftMode.RELATIVE:
            return self._relativeShiftWidget.isFrameBLRFlip()
        else:
            raise NotImplementedError("mode not handled")

    @docstring(_FrameShiftsBase)
    def getFrameBShift(self) -> tuple:
        if self.getShiftMode() is self.ShiftMode.ABSOLUTE:
            return self._absoluteShiftWidget.getFrameBShift()
        elif self.getShiftMode() is self.ShiftMode.RELATIVE:
            return self._relativeShiftWidget.getFrameBShift()
        else:
            raise NotImplementedError("mode not handled")


class Relative2FramesShift(qt.QWidget, _FrameShiftsBase):
    sigShiftsChanged = qt.Signal(tuple)
    """Signal emit when the shift changed. Tuple contains
     shift_imgA, shift_imgB.
     Both are tuple of (x, y, lrflip) x and y are shift values as float.
     lrflip is a boolean notifying if we should flip image or not before applying the shift
    """

    DEFAULT_SINGLE_STEP = 0.5

    MIN_SINGLE_STEP = 0.0001
    MAX_SINGLE_STEP = 1000

    RANGE_MIN_VALUE = -sys.float_info.max
    RANGE_MAX_VALUE = sys.float_info.max

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setLayout(qt.QHBoxLayout())
        # left spacer
        self._lSpacer = qt.QWidget(self)
        self.layout().addWidget(self._lSpacer)
        # define shift info (x and y shift, x-y shift)
        self._infoWidget = qt.QWidget(self)
        self._infoWidget.setLayout(qt.QFormLayout())
        tSpacer = qt.QWidget(self)
        tSpacer.setSizePolicy(
            qt.QSizePolicy.MinimumExpanding, qt.QSizePolicy.MinimumExpanding
        )
        self._infoWidget.layout().addRow(tSpacer)
        self._shiftsWidget = qt.QWidget(self)
        self._shiftsWidget.setLayout(qt.QHBoxLayout())
        self._shiftsWidget.layout().setContentsMargins(0, 0, 0, 0)
        self._infoLabel = qt.QLabel(
            "note: This apply x shift on frame A. Frame B will be left-right flip and a -x shit and y shift are applied",
            self,
        )
        font = self._infoLabel.font()
        font.setPixelSize(14)
        font.setItalic(True)
        self._infoLabel.setFont(font)
        self._infoLabel.setAlignment(qt.Qt.AlignRight | qt.Qt.AlignVCenter)
        self._infoWidget.layout().addRow(self._infoLabel)
        self._infoWidget.layout().addRow("shift", self._shiftsWidget)
        self._xShiftQLE = qt.QDoubleSpinBox(self)
        self._xShiftQLE.setPrefix("x:")
        self._xShiftQLE.setSuffix(" px")
        self._xShiftQLE.setRange(self.RANGE_MIN_VALUE, self.RANGE_MAX_VALUE)
        self._yShiftQLE = qt.QDoubleSpinBox(self)
        self._yShiftQLE.setPrefix("y:")
        self._yShiftQLE.setSuffix(" px")
        self._yShiftQLE.setRange(self.RANGE_MIN_VALUE, self.RANGE_MAX_VALUE)
        self._shiftsWidget.layout().addWidget(self._xShiftQLE)
        self._shiftsWidget.layout().addWidget(self._yShiftQLE)
        self.layout().addWidget(self._infoWidget)
        self._shiftStepSize = qt.QDoubleSpinBox(self)
        self._shiftStepSize.setValue(1)
        self._shiftStepSize.setRange(self.MIN_SINGLE_STEP, self.MAX_SINGLE_STEP)
        self._infoWidget.layout().addRow("steps", self._shiftStepSize)
        bSpacer = qt.QWidget(self)
        bSpacer.setSizePolicy(
            qt.QSizePolicy.MinimumExpanding, qt.QSizePolicy.MinimumExpanding
        )
        self._infoWidget.layout().addRow(bSpacer)

        # define control arrows
        self._controlWidget = _ControlArrowWidget(self)
        self._controlWidget.layout().setContentsMargins(0, 0, 0, 0)
        self.layout().addWidget(self._controlWidget)
        # right spacer
        self._rSpacer = qt.QWidget(self)
        self.layout().addWidget(self._rSpacer)

        # connect signal / slot
        self._controlWidget.sigMoved.connect(self.move)
        self._xShiftQLE.valueChanged.connect(self._emitHasMoved)
        self._yShiftQLE.valueChanged.connect(self._emitHasMoved)
        self._shiftStepSize.valueChanged.connect(self._stepSizeChanged)

        # set up
        self._controlWidget.setFocusPolicy(qt.Qt.StrongFocus)
        self._controlWidget.setFocus(qt.Qt.OtherFocusReason)

    def move(self, direction: str):
        direction = _ControlArrowWidget.Direction.from_value(direction)
        shift = self._shiftStepSize.value()
        if direction is _ControlArrowWidget.Direction.RIGHT:
            self._xShiftQLE.setValue(self._xShiftQLE.value() + shift)
        elif direction is _ControlArrowWidget.Direction.LEFT:
            self._xShiftQLE.setValue(self._xShiftQLE.value() - shift)
        if direction is _ControlArrowWidget.Direction.UP:
            self._yShiftQLE.setValue(self._yShiftQLE.value() + shift)
        elif direction is _ControlArrowWidget.Direction.DOWN:
            self._yShiftQLE.setValue(self._yShiftQLE.value() - shift)
        old = self.blockSignals(True)
        self.blockSignals(old)
        self._emitHasMoved()

    def _emitHasMoved(self):
        self.sigShiftsChanged.emit(
            (
                (
                    self.getFrameAXShift(),
                    self.getFrameAYShift(),
                    self.isFrameALRFlip(),
                ),
                (
                    self.getFrameBXShift(),
                    self.getFrameBYShift(),
                    self.isFrameBLRFlip(),
                ),
            )
        )

    def getFrameAXShift(self):
        return self._xShiftQLE.value()

    def getFrameAYShift(self):
        # to me does not make sense to have a up / down flip
        # neither than having the same shift or an inverted one.
        # so shift will always be apply on frame only B
        return 0

    def getFrameBXShift(self):
        return -self._xShiftQLE.value()

    def getFrameBYShift(self):
        return self._yShiftQLE.value()

    @docstring(_FrameShiftsBase)
    def isFrameALRFlip(self) -> bool:
        return False  # lr flip by default

    @docstring(_FrameShiftsBase)
    def getFrameAShift(self) -> tuple:
        return (
            self.getFrameAXShift(),
            self.getFrameAYShift(),
        )

    @docstring(_FrameShiftsBase)
    def isFrameBLRFlip(self) -> bool:
        return True  # lr flip by default

    @docstring(_FrameShiftsBase)
    def getFrameBShift(self) -> tuple:
        return (
            self.getFrameBXShift(),
            self.getFrameBYShift(),
        )

    def _stepSizeChanged(self, step):
        self._xShiftQLE.setSingleStep(step)
        self._yShiftQLE.setSingleStep(step)

    def keyPressEvent(self, event):
        self._controlWidget.keyPressEvent(event=event)

    def setShiftStep(self, step):
        self._shiftStepSize.setValue(step)


class _ControlArrowWidget(qt.QWidget):
    """Widget with arrows"""

    sigMoved = qt.Signal(str)
    """signal emit when one direction is activated. Will contain the direction in which we want to move"""

    class Direction(_Enum):
        LEFT = "left"
        RIGHT = "right"
        UP = "up"
        DOWN = "down"

    VALID_DIRECTIONS = Direction.members()

    ARROW_BUTTON_SIZE = 30

    QKEY_TO_DIR = {
        qt.Qt.Key_Left: Direction.LEFT,
        qt.Qt.Key_Right: Direction.RIGHT,
        qt.Qt.Key_Up: Direction.UP,
        qt.Qt.Key_Down: Direction.DOWN,
    }

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setLayout(qt.QGridLayout())
        self.layout().setSpacing(2)
        self.layout().setContentsMargins(2, 2, 2, 2)
        style = qt.QApplication.style()
        # left arrow
        self._leftArrow = qt.QPushButton()
        self._leftArrow.setIcon(style.standardIcon(qt.QStyle.SP_ArrowLeft))
        self._leftArrow.setMinimumSize(
            qt.QSize(self.ARROW_BUTTON_SIZE, self.ARROW_BUTTON_SIZE)
        )
        self.layout().addWidget(self._leftArrow, 1, 0, 1, 1)
        # right arrow
        self._rightArrow = qt.QPushButton()
        self._rightArrow.setIcon(style.standardIcon(qt.QStyle.SP_ArrowRight))
        self._rightArrow.setMinimumSize(
            qt.QSize(self.ARROW_BUTTON_SIZE, self.ARROW_BUTTON_SIZE)
        )
        self.layout().addWidget(self._rightArrow, 1, 3, 1, 1)
        # up arrow
        self._upArrow = qt.QPushButton()
        self._upArrow.setIcon(style.standardIcon(qt.QStyle.SP_ArrowUp))
        self._upArrow.setMinimumSize(
            qt.QSize(self.ARROW_BUTTON_SIZE, self.ARROW_BUTTON_SIZE)
        )
        self.layout().addWidget(self._upArrow, 0, 1, 1, 1)
        # down
        self._downArrow = qt.QPushButton()
        self._downArrow.setIcon(style.standardIcon(qt.QStyle.SP_ArrowDown))
        self._downArrow.setMinimumSize(
            qt.QSize(self.ARROW_BUTTON_SIZE, self.ARROW_BUTTON_SIZE)
        )
        self.layout().addWidget(self._downArrow, 3, 1, 1, 1)

        # connect signal / slot
        self._leftArrow.released.connect(self._leftActivated)
        self._rightArrow.released.connect(self._rightActivated)
        self._upArrow.released.connect(self._upActivated)
        self._downArrow.released.connect(self._downActivated)

        # add keyboard shortcut
        self._leftArrowShortCut = qt.QShortcut(qt.Qt.LeftArrow, self)
        self._leftArrowShortCut.activated.connect(self._leftActivated)

    def _leftActivated(self, *args, **kwargs):
        self.sigMoved.emit(self.Direction.LEFT.value)

    def _rightActivated(self, *args, **kwargs):
        self.sigMoved.emit(self.Direction.RIGHT.value)

    def _upActivated(self, *args, **kwargs):
        self.sigMoved.emit(self.Direction.UP.value)

    def _downActivated(self, *args, **kwargs):
        self.sigMoved.emit(self.Direction.DOWN.value)

    def keyPressEvent(self, event):
        key = event.key()
        if key in self.QKEY_TO_DIR:
            self.sigMoved.emit(self.QKEY_TO_DIR[key].value)


class Absolute2FramesShift(qt.QWidget, _FrameShiftsBase):
    sigShiftsChanged = qt.Signal(tuple)
    """Signal emit when the shift changed. Tuple is (shift_imgA, shift_imgB).
     Both are tuple of (x, y, lrflip) x and y are shift values as float.
     lrflip is a boolean notifying if we should flip image or not before applying the shift
    """

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setLayout(qt.QVBoxLayout())
        self._infoLabel = qt.QLabel(
            "note: This apply shift of images and not of the cor. Do not expect to have the same behavior as the Axis widget manual mode",
            self,
        )
        font = self._infoLabel.font()
        font.setPixelSize(14)
        font.setItalic(True)
        self._infoLabel.setFont(font)
        self._infoLabel.setAlignment(qt.Qt.AlignRight | qt.Qt.AlignVCenter)
        self.layout().addWidget(self._infoLabel)

        self._frameAShiftWidget = FrameShiftWidget(title="frame A", parent=self)
        self.layout().addWidget(self._frameAShiftWidget)
        self._frameBShiftWidget = FrameShiftWidget(title="frame B", parent=self)
        self.layout().addWidget(self._frameBShiftWidget)
        spacer = qt.QWidget(self)
        spacer.setSizePolicy(
            qt.QSizePolicy.MinimumExpanding, qt.QSizePolicy.MinimumExpanding
        )
        self.layout().addWidget(spacer)

        # connect signal / slot
        self._frameAShiftWidget.sigShiftChanged.connect(self._shiftHasChanged)
        self._frameBShiftWidget.sigShiftChanged.connect(self._shiftHasChanged)

    def _shiftHasChanged(self, *args, **kwargs):
        self.sigShiftsChanged.emit(
            (
                (
                    self.getFrameAShift()[0],
                    self.getFrameAShift()[1],
                    self.isFrameALRFlip(),
                ),
                (
                    self.getFrameBShift()[0],
                    self.getFrameBShift()[1],
                    self.isFrameBLRFlip(),
                ),
            ),
        )

    @docstring(_FrameShiftsBase)
    def isFrameALRFlip(self) -> bool:
        return self._frameAShiftWidget.isLRFlip()

    @docstring(_FrameShiftsBase)
    def getFrameAShift(self) -> tuple:
        return (
            self._frameAShiftWidget.getXShift(),
            self._frameAShiftWidget.getYShift(),
        )

    @docstring(_FrameShiftsBase)
    def isFrameBLRFlip(self) -> bool:
        return self._frameBShiftWidget.isLRFlip()

    @docstring(_FrameShiftsBase)
    def getFrameBShift(self) -> tuple:
        return (
            self._frameBShiftWidget.getXShift(),
            self._frameBShiftWidget.getYShift(),
        )


class FrameShiftWidget(qt.QWidget):
    sigShiftChanged = qt.Signal(tuple)
    """Signal emit when the shift changed. Tuple contains (x, y, lrflip)
     x, y values as float and lrflip as boolean"""

    DEFAULT_SINGLE_STEP = 0.5

    MIN_SINGLE_STEP = 0.0001
    MAX_SINGLE_STEP = 1000

    RANGE_MIN_VALUE = -sys.float_info.max
    RANGE_MAX_VALUE = sys.float_info.max

    def __init__(self, title, parent=None) -> None:
        super().__init__(parent)
        self.setLayout(qt.QHBoxLayout())
        self.layout().setContentsMargins(0, 0, 0, 0)
        # define label
        self._label = qt.QLabel(f"{title} shift", self)
        self._label.setAlignment(qt.Qt.AlignRight | qt.Qt.AlignVCenter)
        self.layout().addWidget(self._label)
        # define x shift
        self._xShiftQDSB = qt.QDoubleSpinBox(self)
        self._xShiftQDSB.setPrefix("x:")
        self._xShiftQDSB.setSuffix("px")
        self._xShiftQDSB.setValue(0)
        self._xShiftQDSB.setRange(self.RANGE_MIN_VALUE, self.RANGE_MAX_VALUE)
        self.layout().addWidget(self._xShiftQDSB)
        # define y shift
        self._yShiftQDSB = qt.QDoubleSpinBox(self)
        self._yShiftQDSB.setPrefix("y:")
        self._yShiftQDSB.setSuffix("px")
        self._yShiftQDSB.setValue(0)
        self._yShiftQDSB.setRange(self.RANGE_MIN_VALUE, self.RANGE_MAX_VALUE)
        self.layout().addWidget(self._yShiftQDSB)
        # reset button
        self._resetPB = qt.QPushButton(self)
        style = qt.QApplication.style()
        resetIcon = style.standardIcon(qt.QStyle.SP_DialogResetButton)
        self._resetPB.setIcon(resetIcon)
        self._resetPB.setSizePolicy(qt.QSizePolicy.Fixed, qt.QSizePolicy.Fixed)
        self.layout().addWidget(self._resetPB)
        # shift steps
        self._shiftStepsQDSB = qt.QDoubleSpinBox(self)
        self._shiftStepsQDSB.setPrefix("steps:")
        self._shiftStepsQDSB.setSuffix("px")
        self._shiftStepsQDSB.setValue(1.0)
        self._shiftStepsQDSB.setRange(self.MIN_SINGLE_STEP, self.MAX_SINGLE_STEP)
        self.layout().addWidget(self._shiftStepsQDSB)
        self._shiftStepsQDSB.setRange(0.01, 100000)
        self._shiftStepsQDSB.setSingleStep(0.5)
        # lr flip
        self._lrFlipQCB = qt.QCheckBox("lr flip", self)
        self._lrFlipQCB.setToolTip("left-right flip")
        self.layout().addWidget(self._lrFlipQCB)
        self._lrFlipQCB.setChecked(False)

        # connect signal / slot
        self._resetPB.released.connect(self.clear)
        self._xShiftQDSB.valueChanged.connect(self._shiftHasChanged)
        self._yShiftQDSB.valueChanged.connect(self._shiftHasChanged)
        self._shiftStepsQDSB.valueChanged.connect(self._stepsHasChanged)
        self._lrFlipQCB.toggled.connect(self._shiftHasChanged)

    def _stepsHasChanged(self, *args, **kwargs):
        self._xShiftQDSB.setSingleStep(self.getSingleStep())
        self._yShiftQDSB.setSingleStep(self.getSingleStep())

    def getSingleStep(self):
        return self._shiftStepsQDSB.value()

    def getXShift(self) -> float:
        return self._xShiftQDSB.value()

    def getYShift(self) -> float:
        return self._yShiftQDSB.value()

    def clear(self):
        old = self.blockSignals(True)
        self._xShiftQDSB.setValue(0)
        self._yShiftQDSB.setValue(0)
        self.blockSignals(old)
        self._shiftHasChanged()

    def _shiftHasChanged(self, *args, **kwargs):
        shift = self.getXShift(), self.getYShift()
        self.sigShiftChanged.emit(shift)

    def isLRFlip(self) -> bool:
        return self._lrFlipQCB.isChecked()



# coding: utf-8
from __future__ import annotations


from silx.gui import qt
from nxtomo.nxobject.nxdetector import ImageKey

from tomwer.core.scan.nxtomoscan import NXtomoScan
from tomwer.core.utils.nxtomoutils import get_n_series


def check_dark_series(scan: NXtomoScan, logger=None, user_input: bool = True) -> bool:
    """
    check:
        - if scan has computed_dark attached. If there is more than one then ask confirmation to the user to process the dataset
        - else if the NXTomo has more than one series of dark. Otherwise ask confirmation to the user to process the dataset

    ask confirmation for processing is done by a pop up dialog.

    :param scan: scan to check
    :param user_input: if True and if n series of flat is invalid then ask the user to confirm the processing. Else print a warning and refuse processing

    :return: True if processing can be done else False
    :note: user can 'force' the processing to be done
    """
    if not isinstance(scan, NXtomoScan):
        raise TypeError(
            f"scan is expected to be an instance of {NXtomoScan} not a {type(scan)}"
        )
    try:
        image_keys = scan.image_keys
    except AttributeError:
        image_keys = scan.image_key

    n_series = get_n_series(
        image_key_values=image_keys, image_key_type=ImageKey.DARK_FIELD
    )
    if n_series != 1:
        text = f"{n_series} series of dark found when one expected. Nabu won't be able to run reconstruction (flat field will fail)."
        if n_series > 1:
            text += "\nYou can edit image_keys from the `image-key-editor` to ignore a series (invalid all keys of series you want to ignore)."
        if user_input is False:
            text += "\n skip processing"
            if logger is not None:
                logger.error(text)
            return False
        else:
            msg = qt.QMessageBox()
            msg.setIcon(qt.QMessageBox.Warning)
            types = qt.QMessageBox.Ok | qt.QMessageBox.Cancel
            msg.setStandardButtons(types)
            text += "\nContinue ?"
            msg.setText(text)
            return msg.exec_() == qt.QMessageBox.Ok
    else:
        return True


def check_flat_series(scan: NXtomoScan, logger, user_input: bool = True) -> bool:
    """
    Insure the scan contains at least one series of flat. Otherwise warn the user nabu will not be able to process
    """
    if not isinstance(scan, NXtomoScan):
        raise TypeError(
            f"scan is expected to be an instance of {NXtomoScan} not a {type(scan)}"
        )
    try:
        image_keys = scan.image_keys
    except AttributeError:
        image_keys = scan.image_key
    n_series = get_n_series(
        image_key_values=image_keys, image_key_type=ImageKey.FLAT_FIELD
    )
    if n_series == 0:
        text = "No flat found. Nabu will no be able to do the flat field correction and to reconstruct dataset."
        if user_input is False:
            if logger is not None:
                logger.error(text)
            return False
        else:
            msg = qt.QMessageBox()
            msg.setIcon(qt.QMessageBox.Warning)
            types = qt.QMessageBox.Ok | qt.QMessageBox.Cancel
            msg.setStandardButtons(types)
            text += "\nContinue ?"
            msg.setText(text)
            return msg.exec_() == qt.QMessageBox.Ok
    else:
        return True
